theory elimination_stack
imports Main concur_history_relations
begin

context wellformed_history
begin

end

locale elimination_stack = wellformed_history +
  fixes lin :: "'a rel"
  assumes ef_linearizable_def: "sequential_stack lin (elimination_free_val pre val) f"
  assumes pr_in_lin : "(a, b) \<in> elimination_free_pre pre val \<Longrightarrow> (a, b) \<in> lin"
(*  assumes only_ef_def : "(a, b) \<in> lin \<Longrightarrow> a \<in> elimination_free pre val \<and> b \<in> elimination_free pre val"*)
  assumes finite_set_def : "finite (elimination_free pre val)"
  assumes pre_abcd : "abcd pre"

lemma elimination_stackI : "wellformed_history pre val f \<Longrightarrow>
  sequential_stack lin (elimination_free_val pre val) f \<Longrightarrow>
  \<forall> a b . (a, b) \<in> elimination_free_pre pre val \<longrightarrow> (a, b) \<in> lin \<Longrightarrow>
  finite (elimination_free pre val) \<Longrightarrow>
  abcd pre \<Longrightarrow>
  elimination_stack pre val f lin"
apply (unfold elimination_stack_def)
apply (rule conjI)
apply (assumption)
apply (unfold elimination_stack_axioms_def)
apply (rule conjI)+
apply (assumption)+
apply (rule conjI)
by (assumption)+

context elimination_stack
begin

definition same_val :: "'a rel"
  where "same_val = {(a, b) . (a, b) \<in> val \<or> (b, a) \<in> val}"

lemma same_valD : "(a, b) \<in> same_val \<Longrightarrow> (a, b) \<in> val \<or> (b, a) \<in> val"
apply (unfold same_val_def)
by (simp add : set_eq_iff)

lemma same_valI1 : "(a, b) \<in> val \<Longrightarrow> (a, b) \<in> same_val"
apply (unfold same_val_def)
by (simp add : set_eq_iff)

lemma same_valI2 : "(a, b) \<in> val \<Longrightarrow> (b, a) \<in> same_val"
apply (unfold same_val_def)
by (simp add : set_eq_iff)

lemma always_same_val1 : "(a, b) \<in> pre \<Longrightarrow>  \<exists> c . (a, c) \<in> same_val"
by (metis def_push_or_pop same_valI1 same_valI2)

lemma always_same_val2 : "(a, b) \<in> pre \<Longrightarrow>  \<exists> c . (b, c) \<in> same_val"
by (metis def_push_or_pop same_valI1 same_valI2)

lemma not_same_valD : "(a, b) \<in> pre \<Longrightarrow> (a, b) \<notin> same_val \<Longrightarrow> a = b \<or> 
    (\<exists> ra rb . (a, ra) \<in> same_val \<and> (b, rb) \<in> same_val \<and> a \<noteq> b \<and>  a \<noteq> rb \<and> ra \<noteq> b \<and> ra \<noteq> rb)"
by (metis always_same_val1 always_same_val2 def_pop_once def_push_once def_value pop_once_def push_once_def same_valD same_valI1 same_valI2 val_order_def)

lemma same_val_commutative : "(a, b) \<in> same_val \<Longrightarrow> (b, a) \<in> same_val"
by (metis same_valD same_valI1 same_valI2)

lemma same_val_unique : "(a, b) \<in> same_val \<Longrightarrow> (a, c) \<in> same_val \<Longrightarrow> b = c"
apply (drule same_valD)+
apply (erule disjE)+
apply (metis def_pop_once pop_once_def)
apply (metis def_value val_order_def)
apply (erule disjE)
apply (metis def_value val_order_def)
by (metis def_push_once push_once_def)

definition ef :: "'a set"
where
  "ef = {a . (\<exists> b . (a, b) \<in> same_val \<and> ((a, b) \<in> pre \<or> (b, a) \<in> pre))}"

lemma efD : "a \<in> ef \<Longrightarrow> \<exists> b . (a, b) \<in> same_val \<and> ((a, b) \<in> pre \<or> (b, a) \<in> pre)"
apply (unfold ef_def)
by (simp add: set_eq_iff)

lemma efI1 : "(a, b) \<in> same_val \<Longrightarrow> (a, b) \<in> pre \<Longrightarrow> a \<in> ef"
apply (unfold ef_def)
apply (simp add: set_eq_iff)
by metis

lemma efI2 : "(a, b) \<in> same_val \<Longrightarrow> (a, b) \<in> pre \<Longrightarrow> b \<in> ef"
apply (unfold ef_def)
apply (simp add: set_eq_iff)
by (metis same_valD same_valI1 same_valI2)

lemma ef_val_then_ef : "a \<in> ef \<Longrightarrow> (a, b) \<in> same_val \<Longrightarrow> b \<in> ef"
by (metis efD efI1 efI2 same_val_commutative same_val_unique)

lemma ef_then_push_or_pop : "a \<in> ef \<Longrightarrow> is_push a \<or> is_pop a"
by (metis efD is_popI is_pushI same_valD)

definition el :: "'a set"
  where "el = {a . (is_push a \<or> is_pop a) \<and> a \<notin> ef}"

lemma elD : "a \<in> el \<Longrightarrow> \<exists> b . (a, b) \<in> same_val \<and> (a, b) \<notin> pre \<and> (b, a) \<notin> pre"
apply (unfold el_def)
apply (simp add:set_eq_iff)
by (metis efI1 efI2 is_popD lem_push_then_pop_exists same_valI1 same_valI2)

lemma elI1 : "(a, b) \<in> same_val \<Longrightarrow> (a, b) \<notin> pre \<Longrightarrow> (b, a) \<notin> pre \<Longrightarrow> a \<in> el"
by (metis (lifting, no_types) efD el_def lem_val_then_push_and_pop mem_Collect_eq same_valD same_val_unique)

lemma elI2 : "(a, b) \<in> same_val \<Longrightarrow> (a, b) \<notin> pre \<Longrightarrow> (b, a) \<notin> pre \<Longrightarrow> b \<in> el"
by (metis elI1 same_val_commutative)

lemma elI3 : "a \<in> el \<Longrightarrow> (a, b) \<in> same_val \<Longrightarrow> b \<in> el"
by (metis elD elI2 same_val_unique)

lemma el_or_ef1 : "(a, b) \<in> pre \<Longrightarrow> a \<in> el \<or> a \<in> ef"
by (metis always_same_val1 efI1 efI2 elI1 same_val_commutative)

lemma el_or_ef2 : "(a, b) \<in> pre \<Longrightarrow> b \<in> el \<or> b \<in> ef"
by (metis (lifting) el_def lem_push_or_pop2 mem_Collect_eq)

lemma push_then_el_or_ef : "is_push a \<Longrightarrow> a \<in> el \<or> a \<in> ef"
by (metis (lifting) el_def mem_Collect_eq)

lemma pop_then_el_or_ef : "is_pop a \<Longrightarrow> a \<in> el \<or> a \<in> ef"
by (metis (lifting) el_def mem_Collect_eq)

lemma el_then_not_ef : "a \<in> el \<Longrightarrow> a \<notin> ef"
by (metis (lifting, no_types) el_def mem_Collect_eq)

lemma ef_then_not_el : "a \<in> ef \<Longrightarrow> a \<notin> el"
by (metis el_then_not_ef)

definition ef_pre :: "'a rel"
where
  "ef_pre = {(a, b) . a \<in> ef \<and> b \<in> ef \<and> (a, b) \<in> pre}"

lemma ef_preD : "(a, b) \<in> ef_pre \<Longrightarrow> a \<in> ef \<and> b \<in> ef \<and> (a, b) \<in> pre"
apply (unfold ef_pre_def)
by (simp add: set_eq_iff)

lemma ef_preI : "a \<in> ef \<Longrightarrow> b \<in> ef \<Longrightarrow> (a, b) \<in> pre \<Longrightarrow> (a, b) \<in> ef_pre"
apply (unfold ef_pre_def)
by (simp add: set_eq_iff)

definition ef_val :: "'a rel"
where
  "ef_val = {(a, b) . a \<in> ef \<and> b \<in> ef \<and> (a, b) \<in> val}"

lemma ef_valD : "(a, b) \<in> ef_val \<Longrightarrow> a \<in> ef \<and> b \<in> ef \<and> (a, b) \<in> val"
apply (unfold ef_val_def)
by (simp add: set_eq_iff)

lemma ef_valI : "a \<in> ef \<Longrightarrow> b \<in> ef \<Longrightarrow> (a, b) \<in> val \<Longrightarrow> (a, b) \<in> ef_val"
apply (unfold ef_val_def)
by (simp add: set_eq_iff)

lemma ef_val_then_elimination_free_val : "(a, b) \<in> ef_val \<Longrightarrow> (a, b) \<in> elimination_free_val pre val"
by (metis def_val_then_not_pr ef_valD elI1 el_then_not_ef elimination_freeI1 elimination_freeI2 elimination_free_valI sameValI1 same_valI2 val_then_not_pr_def)

lemma elimination_free_val_then_ef_val : "(a, b) \<in> elimination_free_val pre val \<Longrightarrow> (a, b) \<in> ef_val"
by (metis def_pop_once def_val_then_not_pr def_value efI1 efI2 ef_valI elimination_freeD elimination_free_valD pop_once_def sameValD same_valI1 val_order_def val_then_not_pr_def)

lemma ef_val_equals_elimination_free_val : "elimination_free_val pre val = ef_val"
apply (simp add:set_eq_iff)
by (metis ef_val_then_elimination_free_val elimination_free_val_then_ef_val)

lemma ef_linearizable: "sequential_stack lin ef_val f"
by (metis ef_linearizable_def ef_val_equals_elimination_free_val)

(*lemma only_ef : "(a, b) \<in> lin \<Longrightarrow> a \<in> ef \<and> b \<in> ef"
by (metis efI1 efI2 elimination_freeD only_ef_def sameValD same_valI1 same_valI2)*)

lemma finite_set : "finite ef"
by (metis ef_def elimination_free_def finite_set_def sameVal_def same_val_def)


definition minf :: "'a rel"
  where "minf = {(a, b) . \<exists> ra rb . (a, ra) \<in> same_val \<and> (b, rb) \<in> same_val \<and> (min (f a) (f ra)) < (min (f b) (f rb))}"

lemma minfD : "(a, b) \<in> minf \<Longrightarrow>  \<exists> ra rb . (a, ra) \<in> same_val \<and> (b, rb) \<in> same_val \<and> (min (f a) (f ra)) < (min (f b) (f rb))"
apply (unfold minf_def)
by (simp add : set_eq_iff)

lemma minfI : "(a, ra) \<in> same_val \<Longrightarrow> (b, rb) \<in> same_val \<Longrightarrow> (min (f a) (f ra)) < (min (f b) (f rb)) \<Longrightarrow> (a, b) \<in> minf"
apply (unfold minf_def)
apply (simp add : set_eq_iff)
by metis

lemma lem_f_orders : "a \<noteq> b \<Longrightarrow> f a < f b \<or> f b < f a"
by (smt injD operations_countable)

lemma lem_min_f_orders : "a \<noteq> c \<Longrightarrow> a \<noteq> d \<Longrightarrow> b \<noteq> c \<Longrightarrow> b \<noteq> d \<Longrightarrow> 
   (min (f a) (f b) ) < (min (f c) (f d)) \<or> (min (f c) (f d) ) < (min (f a) (f b))"
by (smt injD operations_countable)

lemma always_minf : "(a, ra) \<in> same_val \<Longrightarrow> (b, rb) \<in> same_val \<Longrightarrow> a \<noteq> b \<Longrightarrow> (a, b) \<notin> same_val \<Longrightarrow> (a, b) \<in> minf \<or> (b, a) \<in> minf"

by (metis lem_min_f_orders minfI same_val_commutative same_val_unique)

lemma not_minf_then_same_val : "(a, b) \<notin> minf \<Longrightarrow> (b, a) \<notin> minf \<Longrightarrow> 
  is_push a \<or> is_pop a \<Longrightarrow> is_push b \<or> is_pop b \<Longrightarrow> (a, b) \<in> same_val \<or> a=b"
by (metis always_minf is_popD is_push_def same_valI1 same_valI2)

lemma minf_irreflexive : "irrefl minf"
apply (unfold irrefl_def)
apply (rule allI)
apply (rule notI)
apply (drule minfD)
apply (erule exE)+
apply (erule conjE)+
by (metis less_not_sym same_val_unique)

lemma minf_antisymmetric : "antisym minf"
by (smt antisymI minfD same_val_unique)

lemma minf_transitive : "trans minf"
apply (rule transI)
apply (drule minfD)+
by (metis (hide_lams, no_types) dual_order.strict_trans minfI same_val_unique)

lemma same_val_then_not_minf : "(a, b) \<in> same_val \<Longrightarrow> (a, b) \<notin> minf"
by (metis min.commute minfD not_less_iff_gr_or_eq same_val_commutative same_val_unique)

definition el_pre :: "'a rel"
  where "el_pre = {(a, b) . \<exists> c . (b, c) \<in> same_val \<and> ((a, b) \<in> pre \<or> (a, c) \<in> pre)}"

lemma el_preD1 : "(a, b) \<in> el_pre \<Longrightarrow> \<exists> c . (b, c) \<in> same_val \<and> ((a, b) \<in> pre \<or> (a, c) \<in> pre)"
apply (unfold el_pre_def)
by (simp add : set_eq_iff)

lemma el_preD : "(a, b) \<in> el_pre \<Longrightarrow> (b, c) \<in> same_val \<Longrightarrow> (a, b) \<in> pre \<or> (a, c) \<in> pre"
apply (unfold el_pre_def)
apply (simp add : set_eq_iff)
by (metis def_pop_once def_push_once def_value pop_once_def push_once_def same_valD val_order_def)

lemma el_preI1 : "(b, c) \<in> same_val \<Longrightarrow> (a, b) \<in> pre \<or> (a, c) \<in> pre \<Longrightarrow> (a, b) \<in> el_pre"
apply (unfold el_pre_def)
apply (simp add : set_eq_iff)
by metis

lemma el_preI2 : "(b, c) \<in> same_val \<Longrightarrow> (a, b) \<in> pre \<or> (a, c) \<in> pre \<Longrightarrow> (a, c) \<in> el_pre"
apply (unfold el_pre_def)
apply (simp add : set_eq_iff)
by (metis same_valD same_valI1 same_valI2)

lemma el_pre_abcd : "abcd el_pre"
apply (rule abcdI)
apply (drule el_preD1)+
apply (erule exE)+
apply (erule conjE)+
by (metis abcdD el_preI1 pre_abcd)

lemma lem_pos_exists : "b \<in> el \<Longrightarrow> 
   (\<forall> a . a \<in> ef \<longrightarrow> (a, b) \<notin> el_pre) \<or>
   (\<exists> a . a \<in> ef \<and> (a, b) \<in> el_pre \<and> (\<forall> c . c \<in> ef \<and> (a, c) \<in> lin \<longrightarrow> (c, b) \<notin> el_pre))"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
apply (insert finite_set)
apply (erule finite_induct)
apply (metis emptyE)
apply (erule disjE)
apply (case_tac "(x, b) \<in> el_pre")
apply (rule disjI2)
apply (rule_tac x="x" in exI)
apply (rule conjI)
apply (metis insertI1)
apply (rule conjI)
apply metis
apply (metis insertE lem_irreflexive_then_not_equal)
apply (metis insertE)
apply (erule exE)
apply (erule conjE)+
apply (rule disjI2)
apply (case_tac "(x, b) \<in> el_pre")
apply (case_tac "(x, a) \<in> lin")
apply (metis insertE insertI2 transE)
apply (case_tac "(a, x) \<in> lin")
apply (metis insertCI insertE lem_irreflexive_then_not_equal transE)
apply (case_tac "a= x")
apply metis
apply (metis insert_iff)
by (metis insert_iff)

(* el_pos defines the position where an elimination pair should be placed in the 
   final linearization order. a is the position, b, is one item of the elimination
   pair. *)
definition el_pos :: "'a rel"
  where "el_pos = {(a, b) . a \<in> ef \<and> b \<in> el \<and> (a, b) \<in> el_pre \<and> (\<forall> c . c \<in> ef \<and> (a, c) \<in> lin \<longrightarrow> (c, b) \<notin> el_pre)}"

lemma el_posD : "(a, b) \<in> el_pos \<Longrightarrow> a \<in> ef \<and> b \<in> el \<and> (a, b) \<in> el_pre \<and> (\<forall> c . c \<in> ef \<and> (a, c) \<in> lin \<longrightarrow> (c, b) \<notin> el_pre)"
apply (unfold el_pos_def)
by (simp add : set_eq_iff)

lemma el_posI : "a \<in> ef \<Longrightarrow> b \<in> el \<Longrightarrow> (a, b) \<in> el_pre \<Longrightarrow> (\<forall> c . c \<in> ef \<and> (a, c) \<in> lin \<longrightarrow> (c, b) \<notin> el_pre) \<Longrightarrow> (a, b) \<in> el_pos"
apply (unfold el_pos_def)
by (simp add : set_eq_iff)

lemma lem_is_push_ef_then_is_ef_push : "is_push a \<Longrightarrow> a \<in> ef \<Longrightarrow> wellformed_history.is_push ef_val a"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
by (metis ef_valI ef_val_then_ef is_pushD same_valI1 wellformed_history.is_push_def)

lemma lem_is_pop_ef_then_is_ef_pop : "is_pop a \<Longrightarrow> a \<in> ef \<Longrightarrow> wellformed_history.is_pop ef_val a"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
by (metis ef_valI ef_val_then_ef is_popD same_valI2 wellformed_history.is_pop_def)

lemma one_el_pos : "(a, b) \<in> el_pos \<Longrightarrow> (c, b) \<in> el_pos \<Longrightarrow> a=c"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
apply (drule el_posD)+
apply (erule conjE)+
apply (frule_tac a="a" in ef_then_push_or_pop)
apply (frule_tac a="c" in ef_then_push_or_pop)
apply (erule disjE)+
apply (case_tac "a=c")
apply metis
apply (metis lem_is_push_ef_then_is_ef_push)
apply (metis lem_is_pop_ef_then_is_ef_pop lem_is_push_ef_then_is_ef_push)
by (metis lem_is_pop_ef_then_is_ef_pop lem_is_push_ef_then_is_ef_push)

lemma not_el_posD : "\<not>(\<exists> a . (a, b) \<in> el_pos) \<Longrightarrow> b \<in> el \<Longrightarrow> (\<forall> a . a \<in> ef \<longrightarrow> (a, b) \<notin> el_pre)"
by (metis el_posI lem_pos_exists)

definition ef_el :: "'a rel"
  where "ef_el = {(x, y) . 
(x \<in> el \<and> y \<in> ef \<and> (\<forall> a . a \<in> ef \<longrightarrow> (a, x) \<notin> el_pre) ) \<or>
(x \<in> el \<and> y \<in> ef \<and> (\<exists> a . (a, x) \<in> el_pos \<and> (a, y) \<in> lin)) \<or>
(y \<in> el \<and> x \<in> ef \<and> (x, y) \<in> el_pos) \<or>
(y \<in> el \<and> x \<in> ef \<and> (\<exists> a . (a, y) \<in> el_pos \<and> (x, a) \<in> lin))
 }"

lemma ef_elD : "(x, y) \<in> ef_el \<Longrightarrow> (x \<in> el \<and> y \<in> ef \<and> (\<forall> a . a \<in> ef \<longrightarrow> (a, x) \<notin> el_pre) ) \<or>
(x \<in> el \<and> y \<in> ef \<and> (\<exists> a . (a, x) \<in> el_pos \<and> (a, y) \<in> lin)) \<or>
(y \<in> el \<and> x \<in> ef \<and> (x, y) \<in> el_pos) \<or>
(y \<in> el \<and> x \<in> ef \<and> (\<exists> a . (a, y) \<in> el_pos \<and> (x, a) \<in> lin))"
apply (unfold ef_el_def)
by (simp add : set_eq_iff)

lemma ef_elI1 : "x \<in> el \<Longrightarrow> y\<in> ef \<Longrightarrow> (\<forall> a . a \<in> ef \<longrightarrow> (a, x) \<notin> el_pre) \<Longrightarrow> (x, y) \<in> ef_el"
apply (unfold ef_el_def)
by (simp add : set_eq_iff)

lemma ef_elI2 : "x \<in> el \<Longrightarrow> y\<in> ef \<Longrightarrow> (a, x) \<in> el_pos \<Longrightarrow> (a, y) \<in> lin \<Longrightarrow> (x, y) \<in> ef_el"
apply (unfold ef_el_def)
apply (simp add : set_eq_iff)
by metis

lemma ef_elI3 : "x \<in> el \<Longrightarrow> y \<in> ef \<Longrightarrow> (a, x) \<in> el_pos \<Longrightarrow> (y, a) \<in> lin \<Longrightarrow> (y, x) \<in> ef_el"
apply (unfold ef_el_def)
apply (simp add : set_eq_iff)
by metis

lemma ef_elI4 : "x \<in> el \<Longrightarrow> (y, x) \<in> el_pos \<Longrightarrow> (y, x) \<in> ef_el"
apply (unfold ef_el_def)
apply (simp add : set_eq_iff)
by (metis el_posD)

lemma ef_el_total : "x \<in> el \<Longrightarrow> y \<in> ef \<Longrightarrow> (x, y) \<in> ef_el \<or> (y, x) \<in> ef_el"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (frule lem_pos_exists)
apply (erule disjE)
apply (metis ef_elI1)
apply (erule exE)
apply (erule conjE)+
apply (case_tac "(a, y) \<in> lin")
apply (metis ef_elI2 el_posI)
apply (case_tac "(y, a) \<in> lin")
apply (metis ef_elI3 el_posI)
apply (case_tac "a=y")
apply (metis ef_elI4 el_posI)
by (metis efD lem_is_pop_ef_then_is_ef_pop lem_is_push_ef_then_is_ef_push lem_push_or_pop1 lem_push_or_pop2)

lemma pr_in_ef_el_help : "x \<in> ef \<Longrightarrow> y \<in> el \<Longrightarrow> (x, y) \<in> pre \<Longrightarrow> \<not> (\<exists> a . (a, y) \<in> el_pos) \<Longrightarrow> False"
apply (drule not_el_posD)
apply (assumption)
by (metis elD el_preI1)

lemma pr_in_ef_el_help2 : "x \<in> ef \<Longrightarrow> y \<in> el \<Longrightarrow> (x, y) \<in> pre \<Longrightarrow> (a, y) \<in> el_pos \<Longrightarrow> (x = a) \<or> (x, a) \<in> lin"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (case_tac "x=a")
apply metis
apply (case_tac "(x, a) \<in> lin")
apply metis
apply (case_tac "(a, x) \<in> lin")
apply (metis elD el_posD el_preI1)
by (metis ef_then_push_or_pop el_posD lem_is_pop_ef_then_is_ef_pop lem_is_push_ef_then_is_ef_push)

lemma pr_then_not_rev_pr : "x \<in> el \<Longrightarrow> (x, y) \<in> same_val \<Longrightarrow> (x, a) \<in> pre \<Longrightarrow> (a, y) \<notin> pre"
apply (rule notI)
apply (drule elD)
by (smt pre_transitive pre_irreflexive el_preD el_preI1 irrefl_def transD)

lemma ef_then_elimination_free : "a \<in> ef \<Longrightarrow> a \<in> elimination_free pre val"
by (metis efD ef_valI ef_val_equals_elimination_free_val ef_val_then_ef elimination_free_valD same_valD)

lemma ef_pre_then_elimination_free_pre : "(a, b) \<in> ef_pre \<Longrightarrow> (a, b) \<in> elimination_free_pre pre val"
apply (drule ef_preD)
by (metis ef_then_elimination_free elimination_free_preI)

lemma pr_then_lin : "x \<in> el \<Longrightarrow> y \<in> ef \<Longrightarrow> (x, y) \<in> pre \<Longrightarrow> (a, x) \<in> el_pos \<Longrightarrow> (a, y) \<in> lin"
apply (frule el_posD)
apply (erule conjE)+
apply (frule elD)
apply (erule exE)
apply (erule conjE)+
by (metis (lifting, no_types) pre_transitive abcdD ef_preI ef_pre_then_elimination_free_pre el_preD1 pr_in_lin pre_abcd same_val_unique transD)

lemma pr_in_ef_el1 : "x \<in> ef \<Longrightarrow> y \<in> el \<Longrightarrow> (x, y) \<in> pre \<Longrightarrow> (x, y) \<in> ef_el"
by (metis ef_elI3 ef_elI4 pr_in_ef_el_help pr_in_ef_el_help2)

lemma pr_in_ef_el2 : "x \<in> el \<Longrightarrow> y \<in> ef \<Longrightarrow> (x, y) \<in> pre \<Longrightarrow> (x, y) \<in> ef_el"
by (metis ef_elI1 ef_elI2 not_el_posD pr_then_lin)

lemma ef_el_then_ef_and_el : "(a, b) \<in> ef_el \<Longrightarrow> (a \<in> el \<and> b \<in> ef) \<or> (a \<in> ef \<and> b \<in> el)"
by (smt ef_elD)

lemma ef_el_irreflexive : "irrefl ef_el"
by (metis ef_el_then_ef_and_el ef_then_not_el irrefl_def)

lemma ef_el_antisymmetric : "antisym ef_el"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
apply (rule antisymI)
apply (drule ef_elD)+
by (metis ef_then_not_el el_posD transD)

definition el_el :: "'a rel"
  where "el_el = {(x, y) . x \<in> el \<and> y \<in> el \<and>
    ( (x, y) \<in> val \<or>
      (\<not>(\<exists> a . (a, x) \<in> el_pos) \<and> (\<exists> b . (b, y) \<in> el_pos)) \<or>
      (\<not>(\<exists> a . (a, x) \<in> el_pos) \<and> \<not>(\<exists> b . ( b, y) \<in> el_pos) \<and> (\<exists> c. (c, x) \<notin> el_pre \<and> (c, y) \<in> el_pre)) \<or> 
      (\<not>(\<exists> a . (a, x) \<in> el_pos) \<and> \<not>(\<exists> b . ( b, y) \<in> el_pos) \<and> (\<forall> c. (c, x) \<in> el_pre \<longrightarrow> (c, y) \<in> el_pre) \<and> (x, y) \<in> minf) \<or>
      (\<exists> a b . (a, x) \<in> el_pos \<and> (b, y) \<in> el_pos \<and> (a, b) \<in> lin) \<or>
      (\<exists> a . (a, x) \<in> el_pos \<and> (a, y) \<in> el_pos \<and> (\<exists> c. (a, c) \<in> el_pos \<and> (c, x) \<notin> el_pre \<and> (c, y) \<in> el_pre)) \<or> 
      ((\<exists> a . (a, x) \<in> el_pos \<and> (a, y) \<in> el_pos \<and> (\<forall> c. (a, c) \<in> el_pos \<and> (c, x) \<in> el_pre \<longrightarrow> (c, y) \<in> el_pre)) \<and> (x, y) \<in> minf)
 ) }"

lemma el_elD : "(x, y) \<in> el_el \<Longrightarrow> x \<in> el \<and> y \<in> el \<and>
    ( (x, y) \<in> val \<or>
      (\<not>(\<exists> a . (a, x) \<in> el_pos) \<and> (\<exists> b . (b, y) \<in> el_pos)) \<or>
      (\<not>(\<exists> a . (a, x) \<in> el_pos) \<and> \<not>(\<exists> b . ( b, y) \<in> el_pos) \<and> (\<exists> c. (c, x) \<notin> el_pre \<and> (c, y) \<in> el_pre)) \<or> 
      (\<not>(\<exists> a . (a, x) \<in> el_pos) \<and> \<not>(\<exists> b . ( b, y) \<in> el_pos) \<and> (\<forall> c. (c, x) \<in> el_pre \<longrightarrow> (c, y) \<in> el_pre) \<and> (x, y) \<in> minf) \<or>
      (\<exists> a b . (a, x) \<in> el_pos \<and> (b, y) \<in> el_pos \<and> (a, b) \<in> lin) \<or>
      (\<exists> a . (a, x) \<in> el_pos \<and> (a, y) \<in> el_pos \<and> (\<exists> c. (a, c) \<in> el_pos \<and> (c, x) \<notin> el_pre \<and> (c, y) \<in> el_pre)) \<or> 
      ((\<exists> a . (a, x) \<in> el_pos \<and> (a, y) \<in> el_pos \<and> (\<forall> c. (a, c) \<in> el_pos \<and> (c, x) \<in> el_pre \<longrightarrow> (c, y) \<in> el_pre)) \<and> (x, y) \<in> minf))"
apply (unfold el_el_def)
by (simp add : set_eq_iff)


lemma el_elI1 : "x \<in> el \<Longrightarrow> y \<in> el \<Longrightarrow>(x, y) \<in> val \<Longrightarrow> (x, y) \<in> el_el"
apply (unfold el_el_def)
by (simp add : set_eq_iff)

lemma el_elI2 : "x \<in> el \<Longrightarrow> y \<in> el \<Longrightarrow> \<not>(\<exists> a . (a, x) \<in> el_pos) \<Longrightarrow> (b, y) \<in> el_pos \<Longrightarrow> (x, y) \<in> el_el"
apply (unfold el_el_def)
apply (simp add : set_eq_iff)
by metis

lemma el_elI3 : "x \<in> el \<Longrightarrow> y \<in> el \<Longrightarrow> \<not>(\<exists> a . (a, x) \<in> el_pos) \<Longrightarrow> \<not>(\<exists> b . ( b, y) \<in> el_pos) \<Longrightarrow> (c, x) \<notin> el_pre \<Longrightarrow> (c, y) \<in> el_pre \<Longrightarrow> (x, y) \<in> el_el"
apply (unfold el_el_def)
apply (simp add : set_eq_iff)
by metis

lemma el_elI4 : "x \<in> el \<Longrightarrow> y \<in> el \<Longrightarrow> \<not>(\<exists> a . (a, x) \<in> el_pos) \<Longrightarrow> \<not>(\<exists> b . ( b, y) \<in> el_pos) \<Longrightarrow> (\<forall> c. (c, x) \<in> el_pre \<longrightarrow> (c, y) \<in> el_pre) \<Longrightarrow> (x, y) \<in> minf \<Longrightarrow> (x, y) \<in> el_el"
apply (unfold el_el_def)
by (simp add : set_eq_iff)

lemma el_elI5 : "x \<in> el \<Longrightarrow> y \<in> el \<Longrightarrow> (a, x) \<in> el_pos \<Longrightarrow> (b, y) \<in> el_pos \<Longrightarrow> (a, b) \<in> lin \<Longrightarrow> (x, y) \<in> el_el"
apply (unfold el_el_def)
apply (simp add : set_eq_iff)
by metis

lemma el_elI6 : "x \<in> el \<Longrightarrow> y \<in> el \<Longrightarrow> (a, x) \<in> el_pos \<Longrightarrow> (a, y) \<in> el_pos \<Longrightarrow> (\<exists> c. (a, c) \<in> el_pos \<and> (c, x) \<notin> el_pre \<and> (c, y) \<in> el_pre) \<Longrightarrow> (x, y) \<in> el_el"
apply (unfold el_el_def)
apply (simp add : set_eq_iff)
by metis

lemma el_elI7 : "x \<in> el \<Longrightarrow> y \<in> el \<Longrightarrow> (a, x) \<in> el_pos \<Longrightarrow> (a, y) \<in> el_pos \<Longrightarrow> (\<forall> c. (a, c) \<in> el_pos \<and> (c, x) \<in> el_pre \<longrightarrow> (c, y) \<in> el_pre) \<Longrightarrow> (x, y) \<in> minf \<Longrightarrow> (x, y) \<in> el_el"
apply (unfold el_el_def)
apply (simp add : set_eq_iff)
by metis

lemma el_el_then_el_and_el : "(a, b) \<in> el_el \<Longrightarrow> a \<in> el \<and> b \<in> el"
apply (drule el_elD)
by smt

lemma el_el_total : "x \<in> el \<Longrightarrow> y \<in> el \<Longrightarrow> (x = y) \<or> (x, y) \<in> el_el \<or> (y, x) \<in> el_el"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
apply (case_tac "x=y")
apply metis
apply (case_tac "(x, y) \<in> same_val")
apply (metis el_elI1 same_valD)
apply (case_tac "\<exists> a. (a, x) \<in> el_pos")
apply (case_tac "\<exists> b. (b, y) \<in> el_pos")
apply (erule exE)+
apply (case_tac "a=b")
apply (case_tac "(\<forall> c. (a, c) \<in> el_pos \<and> (c, x) \<in> el_pre \<longrightarrow> (c, y) \<in> el_pre)")
apply (case_tac "(x, y) \<in> minf")
apply (rule disjI2)
apply (rule disjI1)
apply (rule el_elI7)
apply (assumption)+
apply metis
apply (assumption)+
apply (case_tac "(\<forall> c. (b, c) \<in> el_pos \<and> (c, y) \<in> el_pre \<longrightarrow> (c, x) \<in> el_pre)")
apply (case_tac "(y, x) \<in> minf")
apply (rule disjI2)+
apply (rule el_elI7)
apply (assumption)+
apply metis
apply (assumption)+
apply (metis (lifting) el_def mem_Collect_eq not_minf_then_same_val)
apply (metis (lifting, no_types) el_elI6)
apply (metis (lifting, no_types) el_elI6)
apply (drule el_posD)+
apply (erule conjE)+
apply (frule_tac a="a" in ef_then_push_or_pop)
apply (frule_tac a="b" in ef_then_push_or_pop)
apply (erule disjE)+
apply (case_tac "a=b")
apply metis
apply (smt el_elI5 el_posI el_then_not_ef lem_is_push_ef_then_is_ef_push)
apply (drule lem_is_push_ef_then_is_ef_push)
apply (assumption)
apply (drule lem_is_pop_ef_then_is_ef_pop)
apply (assumption)
apply (metis (lifting, no_types) el_elI5 el_posI)
apply (erule disjE)
apply (drule lem_is_push_ef_then_is_ef_push)
apply (assumption)
apply (drule lem_is_pop_ef_then_is_ef_pop)
apply (assumption)
apply (metis (lifting, no_types) el_elI5 el_posI)
apply (case_tac "a=b")
apply metis
apply (drule lem_is_pop_ef_then_is_ef_pop)
apply (assumption)
apply (drule lem_is_pop_ef_then_is_ef_pop)
apply (assumption)
apply (metis (lifting, no_types) el_elI5 el_posI)
apply (metis (lifting, no_types) el_elI2)
by (metis (lifting, no_types) always_minf elD el_elI2 el_elI3 el_elI4)

lemma lem_pr_in_el_el_help : "x \<in> el \<Longrightarrow> y \<in> el \<Longrightarrow> (x, y) \<in> pre \<Longrightarrow> (a, x) \<in> el_pos \<Longrightarrow> 
  (b, y) \<in> el_pos \<Longrightarrow> a = b \<or> (a, b) \<in> lin"
by (metis pre_transitive abcd_def elD el_posD el_preD pr_in_ef_el_help2 pre_abcd trans_def)

lemma lem_pr_in_el_el_help2 : "x \<in> el \<Longrightarrow> y \<in> el \<Longrightarrow> (x, y) \<in> pre \<Longrightarrow> (a, x) \<in> el_pos \<Longrightarrow> 
  \<exists> b .(b, y) \<in> el_pos"
by (metis pre_transitive abcd_def elD el_posD el_preD pr_in_ef_el_help pre_abcd trans_def)

lemma lem_pr_in_el_el : "x \<in> el \<Longrightarrow> y \<in> el \<Longrightarrow> (x, y) \<in> pre \<Longrightarrow> (x, y) \<in> el_el"
apply (case_tac "x=y")
apply (metis pre_irreflexive lem_irreflexive_then_not_equal)
apply (case_tac "(x, y) \<in> same_val")
apply (metis def_val_then_not_pr el_elI1 same_valD val_then_not_pr_def)
apply (case_tac "\<exists> a. (a, x) \<in> el_pos")
apply (case_tac "\<exists> b. (b, y) \<in> el_pos")
apply (metis pre_irreflexive elD el_elI5 el_elI6 el_preD el_preI1 lem_irreflexive_then_not_equal lem_pr_in_el_el_help)
apply (metis lem_pr_in_el_el_help2)
by (metis pre_irreflexive elD el_elI2 el_elI3 el_preD el_preI1 lem_irreflexive_then_not_equal)

lemma lem_position_exists : "(ia, ra) \<in> val \<Longrightarrow> (ia, ra) \<notin> pre \<Longrightarrow> 
  (\<forall> a . a \<in> ef \<longrightarrow> ((a, ia) \<notin> pre \<and> (a, ra) \<notin> pre)) \<or> 
  (\<exists> a . a \<in> ef \<and> ((a, ia) \<in> pre \<or> (a, ra) \<in> pre) \<and> (\<forall> b . b \<in> ef \<and> (a, b) \<in> lin \<longrightarrow> (b, ia) \<notin> pre \<and> (b, ra) \<notin> pre))"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
apply (insert finite_set)
apply (erule finite_induct)
apply (metis empty_iff)
apply (erule disjE)
apply (case_tac "(x, ia) \<in> pre")
apply (rule disjI2)
apply (rule_tac x="x" in exI)
apply (rule conjI)
apply (metis insertCI)
apply (rule conjI)
apply metis
apply (rule allI)
apply (rule impI)
apply (erule_tac x="b" in allE)
apply (metis insertE lem_irreflexive_then_not_equal)
apply (metis insertE lem_irreflexive_then_not_equal)
apply (erule exE)+
apply (erule conjE)+
apply (erule disjE)
apply (smt  insert_iff irrefl_def trans_def)
apply (rule disjI2)
apply (case_tac "(x, ia) \<in> pre")
apply (case_tac "(x, a) \<in> lin")
apply (rule_tac x="a" in exI)
apply (metis antisym_def insert_iff)
apply (case_tac "(a, x) \<in> lin")
apply (rule_tac x="x" in exI)
apply (metis (full_types) insert_iff irrefl_def transD)
apply (metis (full_types) insert_iff)
apply (case_tac "(x, ra) \<in> pre")
apply (case_tac "(x, a) \<in> lin")
apply (rule_tac x="a" in exI)
apply (metis antisym_def insert_iff)
apply (case_tac "(a, x) \<in> lin")
apply (rule_tac x="x" in exI)
apply (metis (full_types) insert_iff irrefl_def transD)
apply (metis (full_types) insert_iff)
apply (rule_tac x="a" in exI)
by (metis (full_types) insert_iff)

lemma el_el_irreflexive : "irrefl el_el"
apply (insert ef_linearizable)
apply (unfold irrefl_def)
apply (rule allI)
apply (rule notI)
apply (drule el_elD)
by (metis  def_value el_posD lem_irreflexive_then_not_equal minf_irreflexive val_order_def)

lemma el_el_antisymmetric : "antisym el_el"
apply (insert ef_linearizable)
apply (rule antisymI)
apply (drule el_elD)+
apply (erule conjE)+
apply (erule disjE)+
apply (metis def_value val_order_def)
apply (smt el_posD el_preD el_preI2 not_el_posD same_valI1 same_valI2 same_val_then_not_minf)
apply (erule disjE)
apply (smt el_posD el_posI el_preD el_preI2 same_valI1 same_valI2 same_val_then_not_minf)
apply (erule disjE)
apply smt
apply (erule disjE)
apply smt
apply (erule disjE)+
apply (metis abcd_def el_pre_abcd)
apply smt
apply (erule disjE)
apply smt
apply (erule disjE)+
apply (metis antisymD minf_antisymmetric)
apply smt
apply (erule disjE)
apply metis
apply (erule disjE)+
apply (metis abcdD el_posD el_pre_abcd)
apply (metis el_posD one_el_pos)
apply (erule disjE)
apply (metis el_posD one_el_pos)
apply (erule disjE)
apply (metis abcdD el_posD el_pre_abcd one_el_pos)
apply (erule disjE)
apply (metis one_el_pos)
by (metis antisymD minf_antisymmetric)

lemma val_not_el_pos_then_not_el_pos : "(x, y) \<in> same_val \<Longrightarrow> \<not> (\<exists>a. (a, y) \<in> el_pos) \<Longrightarrow> \<not> (\<exists>a. (a, x) \<in> el_pos)"
by (metis elD elI2 el_posD el_preD el_preI2 not_el_posD same_val_unique)

lemma val_minf_then_minf : "(x, y) \<in> same_val \<Longrightarrow> (y, z) \<in> minf \<Longrightarrow> (x, z) \<in> minf"
by (metis min.commute minfD minfI same_val_commutative same_val_unique)

lemma val_minf_then_minf2 : "(x, y) \<in> same_val \<Longrightarrow> (z, y) \<in> minf \<Longrightarrow> (z, x) \<in> minf"
by (metis min.commute minfD minfI same_val_commutative same_val_unique)

lemma val_then_el_pos1 : "(x, y) \<in> same_val \<Longrightarrow> (a, x) \<in> el_pos \<Longrightarrow> (a, y) \<in> el_pos"
by (smt elI3 el_posD el_posI el_preD el_preI2 same_val_commutative)

lemma val_then_el_pos2 : "(x, y) \<in> same_val \<Longrightarrow> (a, y) \<in> el_pos \<Longrightarrow> (a, x) \<in> el_pos"
by (smt elI3 el_posD el_posI el_preD el_preI2 same_val_commutative)

lemma val_then_el_pre : "(x, y) \<in> same_val \<Longrightarrow> (a, x) \<in> el_pre \<Longrightarrow> (a, y) \<in> el_pre"
by (metis el_preD el_preI2)

lemma el_el_transitive_help : "(y, z) \<in> same_val \<Longrightarrow> (\<forall>c. (a, c) \<in> el_pos \<and> (c, x) \<in> el_pre \<longrightarrow> (c, y) \<in> el_pre) \<Longrightarrow>
  (\<forall>c. (a, c) \<in> el_pos \<and> (c, x) \<in> el_pre \<longrightarrow> (c, z) \<in> el_pre)"
by (metis val_then_el_pre)

lemma el_el_transitive : "trans el_el"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
apply (rule transI)
apply (drule el_elD)+
apply (erule conjE)+
apply (erule disjE)+
apply (metis def_value val_order_def)
apply (erule disjE)
apply (metis (hide_lams, no_types) el_elI2 same_valI2 val_then_el_pos2)
apply (erule disjE)
apply (smt el_elI3 el_posD el_preD el_preI2 not_el_posD same_valI1)
apply (erule disjE)
apply (erule conjE)+
apply (frule same_valI1)
apply (frule val_not_el_pos_then_not_el_pos)
apply (assumption)
apply (frule_tac x="x" and y="y" and z="z" in val_minf_then_minf)
apply (assumption)
apply (metis (hide_lams, no_types) el_elI4 el_preD el_preI2)
apply (erule disjE)
apply (metis (hide_lams, no_types) el_elI5 same_valI1 val_then_el_pos2)
apply (erule disjE)
apply (smt el_elI6 el_preD el_preI1 same_valI1 same_valI2 val_then_el_pos1)
apply (smt el_elI7 el_preD el_preI1 same_valI1 same_valI2 val_minf_then_minf val_then_el_pos2)
apply (erule disjE)+
apply (metis (hide_lams, no_types) el_elI2 same_valI1 val_then_el_pos1)
apply (erule disjE)
apply (metis (hide_lams, no_types) antisym_def el_elI1 el_elI2 el_elI3 el_el_antisymmetric)
apply (erule disjE)
apply (smt el_elI4 el_preD el_preI1 same_valI1 same_valI2 val_minf_then_minf2 val_then_el_pos1)
apply (erule disjE)
apply (metis (hide_lams, no_types) el_elI5 same_valI1 val_then_el_pos1)
apply (erule disjE)
apply (smt el_elI6 el_preD1 el_preI2 same_valI1 same_val_unique val_then_el_pos1)
apply (erule conjE)+
apply (erule exE)
apply (erule conjE)+
apply (frule same_valI1)
apply (frule same_val_commutative)
apply (frule_tac x="z" and y="y" and z="x" in val_minf_then_minf2)
apply (assumption)
apply (frule_tac y="y" and z="z" and x="x" and a="a" in el_el_transitive_help)
apply (assumption)
apply (frule_tac x="y" and y="z" and a="a" in val_then_el_pos1)
apply (assumption)
apply (metis (hide_lams, no_types) el_elI7)
apply (erule disjE)
apply (smt el_elI2)
apply (erule disjE)
apply (smt el_elI2)
apply (erule disjE)
apply (smt abcd_def el_elI3 el_pre_abcd)
apply (erule disjE)
apply (smt el_elI3)
apply (erule disjE)
apply (smt el_elI4 minf_transitive trans_def)
apply (erule disjE)
apply metis
apply (erule disjE)
apply (metis (full_types) el_elI5 one_el_pos transE)
apply (erule disjE)
apply (metis (hide_lams, no_types) el_elI5 one_el_pos)
apply (erule disjE)
apply (smt abcd_def el_elI6 el_pre_abcd one_el_pos)
apply (erule disjE)
apply (metis (hide_lams, no_types) el_elI6 one_el_pos)
apply (erule conjE)+
apply (erule exE)+
apply (rule_tac x="x" and y="z" and a="a" in el_elI7)
apply (assumption)+
apply metis
apply (metis one_el_pos)
apply (metis one_el_pos)
by (metis minf_transitive transD)

lemma ef_el_ef_then_lin : "x \<in> ef \<Longrightarrow> y \<in> el \<Longrightarrow> z \<in> ef \<Longrightarrow> (x, y) \<in> ef_el \<Longrightarrow> (y, z) \<in> ef_el \<Longrightarrow> (x, z) \<in> lin"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
apply (drule ef_elD)+
by (smt ef_then_not_el el_posD one_el_pos transD)

lemma el_ef_el_then_lin : "x \<in> el \<Longrightarrow> y \<in> ef \<Longrightarrow> z \<in> el \<Longrightarrow> (x, y) \<in> ef_el \<Longrightarrow> (y, z) \<in> ef_el \<Longrightarrow> (x, z) \<in> el_el"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
apply (drule ef_elD)+
by (smt el_elI2 el_elI5 el_posD el_then_not_ef trans_def)

lemma el_el_ef_then_lin : "x \<in> el \<Longrightarrow> y \<in> el \<Longrightarrow> z \<in> ef \<Longrightarrow> (x, y) \<in> el_el \<Longrightarrow> (y, z) \<in> ef_el \<Longrightarrow> (x, z) \<in> ef_el"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
apply (drule el_elD)
apply (drule ef_elD)
apply (erule conjE)+
apply (erule disjE)+
apply (metis (hide_lams, no_types) ef_elI1 same_valI1 val_then_el_pre)
apply (erule disjE)
apply (metis el_posD)
apply (smt abcdD ef_elI1 el_posD el_pre_abcd)
apply (erule disjE)
apply (metis (hide_lams, no_types) ef_elI2 ef_then_not_el same_valI2 val_then_el_pos1)
apply (erule disjE)+
apply (metis (hide_lams, no_types) ef_elI1 not_el_posD)
apply (erule disjE)
apply metis
apply (smt ef_elI2 one_el_pos trans_def)
apply (erule disjE)+
apply (metis ef_then_not_el)
apply (metis ef_then_not_el)
by (metis ef_then_not_el)

lemma ef_el_el_then_lin : "x \<in> ef \<Longrightarrow> y \<in> el \<Longrightarrow> z \<in> el \<Longrightarrow> (x, y) \<in> ef_el \<Longrightarrow> (y, z) \<in> el_el \<Longrightarrow> (x, z) \<in> ef_el"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
apply (drule el_elD)
apply (drule ef_elD)
apply (erule disjE)
apply (metis (hide_lams, no_types) el_then_not_ef)
apply (erule disjE)
apply (metis ef_then_not_el)
apply (erule disjE)
apply (smt ef_elI3 ef_elI4 one_el_pos same_valI1 val_then_el_pos1)
apply (erule conjE)+
apply (erule disjE)
apply (metis (hide_lams, no_types) ef_elI3 same_valI1 val_then_el_pos1)
apply (erule disjE)
apply metis
by (smt ef_elI2 ef_elI3 ef_el_ef_then_lin el_posD one_el_pos)

definition ef_lin :: "'a rel"
 where "ef_lin = {(a, b) . (a, b) \<in> lin \<and> a \<in> ef \<and> b \<in> ef}"

lemma ef_linI : "(a, b) \<in> lin \<Longrightarrow> a \<in> ef \<Longrightarrow> b \<in> ef \<Longrightarrow> (a, b) \<in> ef_lin"
apply (unfold ef_lin_def)
by (simp add: set_eq_iff)

lemma ef_linD : "(a, b) \<in> ef_lin \<Longrightarrow> (a, b) \<in> lin \<and> a \<in> ef \<and> b \<in> ef"
apply (unfold ef_lin_def)
by (simp add: set_eq_iff)

lemma ef_lin_transitive: "trans ef_lin"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
by (metis ef_linD ef_linI trans_def)


definition witness :: "'a rel"
  where "witness = ef_lin Un ef_el Un el_el"

lemma lin_then_witness : "(a, b) \<in> ef_lin \<Longrightarrow> (a, b) \<in> witness"
by (metis UnCI witness_def)

lemma ef_el_then_witness : "(a, b) \<in> ef_el \<Longrightarrow> (a, b) \<in> witness"
by (metis UnCI witness_def)

lemma el_el_then_witness : "(a, b) \<in> el_el \<Longrightarrow> (a, b) \<in> witness"
by (metis UnCI witness_def)

lemma pr_in_witness : "(a, b) \<in> pre \<Longrightarrow> (a, b) \<in> witness"
by (metis ef_el_then_witness ef_linI ef_then_elimination_free el_el_then_witness el_or_ef1 el_or_ef2 elimination_free_preI lem_pr_in_el_el lin_then_witness pr_in_ef_el1 pr_in_ef_el2 pr_in_lin)

lemma ef_then_in_lin : "a \<in> ef \<Longrightarrow> b \<in> ef \<Longrightarrow> a \<noteq> b \<Longrightarrow> (a, b) \<in> lin \<or> (b, a) \<in> lin"
by (metis ef_linearizable ef_then_push_or_pop lem_is_pop_ef_then_is_ef_pop lem_is_push_ef_then_is_ef_push sequential_set.total_on_pop sequential_set.total_on_push sequential_set.total_on_push_pop sequential_stackD)

lemma ef_el_b_ef_then_a_el : "(a, b) \<in> ef_el \<Longrightarrow> b \<in> ef \<Longrightarrow> a \<in> el"
by (metis ef_el_then_ef_and_el ef_then_not_el)

lemma witness_transitive : "trans witness"
apply (rule transI)
apply (unfold witness_def)
apply (erule UnE)+
apply (metis ef_lin_transitive lin_then_witness transE witness_def)
apply (drule ef_elD)
apply (erule disjE)
apply (erule conjE)+
apply (metis ef_linD ef_then_not_el)
apply (metis (lifting) Un_def ef_elI3 ef_linD ef_linI ef_lin_transitive el_def el_posD mem_Collect_eq trans_def)
apply (erule UnE)
apply (drule ef_linD)
apply (erule conjE)+
apply (frule ef_el_b_ef_then_a_el)
apply (assumption)
apply (rule UnI1)
apply (rule UnI2)
apply (drule ef_elD)
apply (erule disjE)
apply (metis ef_elI1)
apply (erule disjE)
apply (metis ef_elI2 ef_linD ef_linI ef_lin_transitive el_posD transE)
apply (erule disjE)
apply (metis ef_then_not_el)
apply (metis ef_then_not_el)
apply (metis ef_el_ef_then_lin ef_el_then_ef_and_el ef_linI el_ef_el_then_lin el_el_then_witness el_then_not_ef lin_then_witness witness_def)
apply (erule UnE)
apply (metis ef_linD el_el_then_el_and_el el_then_not_ef)
apply (metis ef_el_el_then_lin ef_el_then_ef_and_el ef_el_then_witness el_el_then_el_and_el el_then_not_ef witness_def)
apply (erule UnE)+
apply (metis ef_linD el_el_then_el_and_el el_then_not_ef)
apply (metis ef_el_then_ef_and_el ef_el_then_witness el_el_ef_then_lin el_el_then_el_and_el el_then_not_ef witness_def)
by (metis el_el_then_witness el_el_transitive trans_def witness_def)

lemma Un_irrefl : "irrefl A \<Longrightarrow> irrefl B \<Longrightarrow> irrefl (A Un B)"
apply (unfold irrefl_def)
apply (rule allI)
apply (rule notI)
by (metis UnE)

lemma witness_irreflexive : "irrefl witness"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
apply (unfold irrefl_def)
apply (rule allI)
apply (rule notI)
apply (unfold witness_def)
apply (erule UnE)+
apply (metis ef_linD)
apply (metis ef_el_irreflexive lem_irreflexive_then_not_equal)
by (metis el_el_irreflexive lem_irreflexive_then_not_equal)

lemma witness_antisymmetric : "antisym witness"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
apply (rule antisymI)
apply (unfold witness_def)
apply (erule UnE)+
apply (metis antisymD ef_linD)
apply (metis ef_el_b_ef_then_a_el ef_linD el_then_not_ef)
apply (metis UnE antisymD ef_el_antisymmetric ef_el_then_ef_and_el ef_linD ef_then_not_el)
apply (erule UnE)
apply (metis ef_linD ef_then_not_el el_el_then_el_and_el)
apply (metis ef_el_then_ef_and_el el_el_then_el_and_el el_then_not_ef)
by (metis el_el_then_witness lem_irreflexive_then_not_equal trans_def witness_def witness_irreflexive witness_transitive)

lemma witness_total_on_push : "is_push a \<Longrightarrow> is_push b \<Longrightarrow> a \<noteq> b \<Longrightarrow> (a, b) \<in> witness \<or> (b, a) \<in> witness"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
apply (frule_tac a="a" in push_then_el_or_ef)
apply (frule_tac a="b" in push_then_el_or_ef)
apply (erule disjE)+
apply (metis el_el_then_witness el_el_total)
apply (metis ef_el_then_witness ef_el_total)
by (metis ef_el_then_witness ef_el_total ef_linI ef_then_in_lin lin_then_witness)

lemma witness_total_on_pop : "is_pop a \<Longrightarrow> is_pop b \<Longrightarrow> a \<noteq> b \<Longrightarrow> (a, b) \<in> witness \<or> (b, a) \<in> witness"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
apply (frule_tac a="a" in pop_then_el_or_ef)
apply (frule_tac a="b" in pop_then_el_or_ef)
apply (erule disjE)+
apply (metis el_el_then_witness el_el_total)
apply (metis ef_el_then_witness ef_el_total)
by (metis ef_el_then_witness ef_el_total ef_linI ef_then_in_lin lin_then_witness)

lemma witness_total_on_push_pop : "is_push a \<Longrightarrow> is_pop b \<Longrightarrow> (a, b) \<in> witness \<or> (b, a) \<in> witness"
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
apply (frule_tac a="a" in push_then_el_or_ef)
apply (frule_tac a="b" in pop_then_el_or_ef)
apply (erule disjE)+
apply (metis def_value el_el_then_witness el_el_total is_popD is_pushD val_order_def)
apply (metis ef_el_then_witness ef_el_total)
by (metis ef_el_then_witness ef_el_total ef_linI lem_is_pop_ef_then_is_ef_pop lem_is_push_ef_then_is_ef_push lin_then_witness)

lemma el_then_same_witness : "x \<in> el \<Longrightarrow> (x, y) \<in> same_val \<Longrightarrow> (a, x) \<in> witness \<Longrightarrow> (x, a) \<in> same_val \<or> (a, y) \<in> witness"
apply (unfold witness_def)
apply (erule UnE)+
apply (metis ef_linD el_then_not_ef)
apply (drule ef_elD)
apply (erule disjE)
apply (metis el_then_not_ef)
apply (erule disjE)
apply (metis ef_then_not_el)
apply (erule disjE)
apply (metis ef_elI4 ef_el_then_witness elI3 val_then_el_pos1 witness_def)
apply (metis ef_elI3 ef_el_then_witness elI3 val_then_el_pos1 witness_def)
apply (drule el_elD)
apply (erule conjE)+
apply (erule disjE)
apply (metis same_valI2)
apply (erule disjE)
apply (metis el_elI2 el_el_then_witness el_posD val_then_el_pos1 witness_def)
apply (erule disjE)
apply (metis elI3 el_elI2 el_elI3 el_el_then_witness val_then_el_pre witness_def)
apply (erule disjE)
apply (erule conjE)+
apply (frule same_val_commutative)
apply (frule_tac x="y" and y="x" and z="a" in val_minf_then_minf2)
apply (assumption)
apply (smt elI3 el_elI4 el_el_then_witness val_then_el_pos1 val_then_el_pre witness_def)
apply (erule disjE)
apply (metis elI3 el_elI5 el_el_then_witness val_then_el_pos1 witness_def)
apply (erule disjE)
apply (smt elI3 el_elI6 el_el_then_witness el_el_transitive_help val_then_el_pos1 witness_def)
apply (erule conjE)+
apply (frule same_val_commutative)
apply (frule_tac x="y" and y="x" and z="a" in val_minf_then_minf2)
apply (assumption)
apply (erule exE)
apply (erule conjE)+
apply (frule_tac x="x" and y="y" and a="aa" in val_then_el_pos1)
apply (assumption)
by (smt elI3 el_elI7 el_el_then_witness minfD same_val_commutative same_val_unique val_then_el_pre witness_def)

lemma witness_then_push_or_pop1 : "(a, b) \<in> witness \<Longrightarrow> is_push a \<or> is_pop a"
by (smt UnE ef_elD ef_linD ef_then_push_or_pop elD el_el_then_el_and_el lem_val_then_push_and_pop same_valD witness_def)

lemma witness_then_push_or_pop2 : "(a, b) \<in> witness \<Longrightarrow> is_push b \<or> is_pop b"
apply (unfold witness_def)
apply (erule UnE)+
apply (metis ef_linD ef_then_push_or_pop)
apply (metis ef_el_then_ef_and_el ef_then_push_or_pop elD lem_val_then_push_and_pop same_valD)
by (metis elD el_el_then_el_and_el lem_val_then_push_and_pop same_valD)

lemma el_then_same_witness2 : "x \<in> el \<Longrightarrow> (x, y) \<in> same_val \<Longrightarrow> (x, a) \<in> witness \<Longrightarrow> (x, a) \<in> same_val \<or> (y, a) \<in> witness"
apply (rotate_tac -1)
apply (erule contrapos_pp)
apply (safe)
apply (case_tac "(a, y) \<in> witness")
apply (metis antisymD elI3 el_then_same_witness lem_irreflexive_then_not_equal same_val_commutative same_val_unique witness_antisymmetric witness_irreflexive)
by (metis is_popI is_pushI same_valD witness_then_push_or_pop2 witness_total_on_pop witness_total_on_push witness_total_on_push_pop)

lemma ef_ef_then_lin : "a \<in> ef \<Longrightarrow> b \<in> ef \<Longrightarrow> (a, b) \<in> witness \<Longrightarrow> (a, b) \<in> ef_lin"
apply (unfold witness_def)
apply (erule UnE)+
apply (assumption)
apply (metis ef_el_b_ef_then_a_el el_then_not_ef)
by (metis el_el_then_el_and_el el_then_not_ef)

lemma ef_val_ef : "(a, b) \<in> val \<Longrightarrow> a \<in> ef \<Longrightarrow> b \<in> ef"
by (metis ef_val_then_ef same_valI1)

lemma witness_ooo : "sequential_lifo witness val"
apply (unfold sequential_lifo_def)
apply (insert ef_linearizable)
apply (unfold sequential_stack_def)
apply (unfold sequential_stack_axioms_def)
apply (erule conjE)+
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold sequential_set_axioms_def)
apply (unfold wellformed_history_def)
apply (erule conjE)
apply (rule allI)+
apply (rule impI)+
apply (frule_tac ia="ia" in is_pushI)
apply (drule push_then_el_or_ef)
apply (erule disjE)
apply (frule_tac x="ia" and y="ra" and a="ib" in  el_then_same_witness2)
apply (metis same_valI1)
apply assumption
apply (erule disjE)
apply (metis def_value same_valD val_order_def)
apply (metis antisymD def_value val_order_def witness_antisymmetric)
apply (frule_tac ia="ib" in is_pushI)
apply (drule push_then_el_or_ef)
apply (erule disjE)
apply (frule_tac x="ib" and y="rb" and a="ra" in el_then_same_witness2)
apply (metis same_valI1)
apply assumption
apply (erule disjE)
apply (metis elI3 el_then_not_ef same_valI2)
apply metis
apply (frule_tac a="ia" and b="ra" in ef_val_then_ef)
apply (metis same_valI1)
apply (frule_tac a="ia" and b="ib" in ef_ef_then_lin)
apply (assumption)+
apply (frule_tac a="ib" and b="ra" in ef_ef_then_lin)
apply (assumption)+
apply (rule lin_then_witness)
apply (frule_tac a="ib" and b="rb" in ef_val_ef)
apply (assumption)
by (smt ef_linD ef_linI ef_valI sequential_lifo_def)

lemma val_then_not_witness : "(ia, ra) \<in> val \<Longrightarrow> (ra, ia) \<notin> witness"
apply (frule_tac ia="ia" in is_pushI)
apply (drule push_then_el_or_ef)
apply (erule disjE)
apply (frule_tac y="ra" in el_elI1)
apply (metis elI3 same_valI1)
apply (assumption)
apply (drule el_el_then_witness)
apply (rule notI)
apply (metis antisymD def_value val_order_def witness_antisymmetric)
by (metis antisym_def def_val_then_not_pr elI2 el_then_not_ef pr_in_witness same_valI2 val_then_not_pr_def witness_antisymmetric)

lemma witness_wellformed_history : "wellformed_history witness val f"
apply (rule wellformed_historyI)
apply (metis witness_transitive)
apply (metis witness_antisymmetric)
apply (metis witness_irreflexive)
apply (metis operations_countable)
apply (metis operations_finite)
apply (metis def_push_once)
apply (metis def_pop_once)
apply (metis def_value)
apply (metis is_pop_def is_push_def witness_then_push_or_pop1 witness_then_push_or_pop2)
by (metis val_then_not_pr_def val_then_not_witness)

lemma witness_sequential_history : "sequential_set witness val f"
apply (unfold sequential_set_def)
apply (rule conjI)
apply (rule witness_wellformed_history)
apply (unfold sequential_set_axioms_def)
apply (rule conjI)
apply (metis witness_total_on_push)
apply (rule conjI)
apply (metis witness_total_on_pop)
by (metis witness_total_on_push_pop)

lemma witness_sequential_stack : "sequential_stack witness val f"
apply (unfold sequential_stack_def)
apply (rule conjI)
apply (rule witness_sequential_history)
apply (unfold sequential_stack_axioms_def)
by (rule witness_ooo)

theorem elimination_stack_linearizable : "\<exists> prt .( pre \<subseteq> prt \<and> sequential_stack prt val f)"
apply (rule_tac x="witness" in exI)
by (metis pr_in_witness subrelI witness_sequential_stack)

end



theorem help_elimination_stack : "wellformed_history pre val f \<Longrightarrow> abcd pre \<Longrightarrow> 
  linearizable_non_empty_stack (elimination_free_pre pre val) (elimination_free_val pre val)
    \<Longrightarrow> \<exists> lin . elimination_stack pre val f lin"
apply (unfold linearizable_non_empty_stack_def)
apply (erule conjE)+
apply (erule exE)+
apply (rule_tac x="prt" in exI)
apply (rule elimination_stackI)
apply (assumption)+
apply (erule conjE)
apply (frule sequential_stackD)
apply (rule sequential_stackI)
apply (rule sequential_setI)
apply (rule wellformed_historyI)
apply (erule conjE)
apply (drule sequential_setD)
apply (erule conjE)+
apply (drule_tac pre="prt" in wellformed_historyD)
apply (erule conjE)+
apply (assumption)+
apply (drule wellformed_historyD)
apply (erule conjE)+
apply (drule sequential_setD)
apply (erule conjE)+
apply (drule_tac pre="prt" in wellformed_historyD)
apply (erule conjE)+
apply (assumption)
apply (metis sequential_setD wellformed_history.pre_irreflexive)
apply (metis wellformed_history.operations_countable)
apply (metis sequential_setD wellformed_history.operations_finite)
apply (metis elimination_free_valD push_once_def wellformed_history.def_push_once)
apply (metis elimination_free_valD pop_once_def wellformed_history.def_pop_once)
apply (metis elimination_free_valD val_order_def wellformed_history.def_value)
apply (erule conjE)+
apply (drule sequential_setD)
apply (erule conjE)+
apply (drule wellformed_historyD)
apply (metis (full_types) wellformed_historyD)
apply (erule conjE)+
apply (drule wellformed_historyD)
apply (metis sequential_setD wellformed_historyD)
apply (metis sequential_setD)
apply (metis sequential_setD)
apply (metis sequential_setD)
apply metis
apply (metis in_mono)
apply (metis finite_elimination_free wellformed_historyD)
by metis


(* theorem help1 : "elimination_stack pre val lin f \<Longrightarrow> linearizable_non_empty_stack pre val" 
sledgehammer
by (metis elimination_stack.elimination_stack_linearizable linearizable_non_empty_stack_def)*)

theorem elimination_theorem : "wellformed_history pre val f \<Longrightarrow> abcd pre \<Longrightarrow>
  linearizable_non_empty_stack (elimination_free_pre pre val) (elimination_free_val pre val) \<Longrightarrow>
  linearizable_non_empty_stack pre val"
apply (drule linearizable_non_empty_stackD)
apply (erule conjE)
apply (erule exE)+
apply (erule conjE)
apply (frule_tac pre="pre" and val="val" in help_elimination_stack)
apply (assumption)+
apply (frule wellformed_historyD)
apply (erule conjE)+
apply (rule_tac pre="(elimination_free_pre pre val)" and prt="prt" and f="fa" and val="(elimination_free_val pre val)" in linearizable_non_empty_stackI)
apply (rule conc_historyI)
apply (metis conc_historyD)
apply (metis conc_historyD)
apply (metis conc_historyD)
apply (rule abcdI)
apply (drule elimination_free_preD)+
apply (erule conjE)+
apply (metis (full_types) abcd_def elimination_free_preI)
apply metis
apply metis
by (metis conc_historyI elimination_stack.elimination_stack_linearizable linearizable_non_empty_stack_def wellformed_history.pre_antisymmetric wellformed_history.pre_irreflexive wellformed_history.pre_transitive)

end
