theory empty_stack
imports Main concur_history_relations

begin
locale concurrent_empty_stack = wellformed_empty_history +
  fixes slin :: "'a rel"
  fixes elin :: "'a rel"
  assumes stack_linearizable: "sequential_stack slin val f"
  assumes pr_in_slin : "(a, b) \<in> pre \<Longrightarrow> \<not> emp a \<Longrightarrow> \<not> emp b \<Longrightarrow> (a, b) \<in> slin"
  assumes set_linearizable: "sequential_empty_set elin val emp f"
  assumes pr_in_elin : "(a, b) \<in> pre \<Longrightarrow> (a, b) \<in> elin"

context wellformed_empty_history
begin

lemma concurrent_empty_stackI : 
"wellformed_empty_history pre val emp f \<Longrightarrow>
 sequential_stack slin val f \<Longrightarrow>
 \<forall> a b . (a, b) \<in> pre \<longrightarrow> \<not> emp a \<longrightarrow> \<not> emp b \<longrightarrow> (a, b) \<in> slin \<Longrightarrow>
 sequential_empty_set elin val emp f \<Longrightarrow>
 \<forall> a b . (a, b) \<in> pre \<longrightarrow> (a, b) \<in> elin \<Longrightarrow>
 concurrent_empty_stack pre val emp f slin elin"
apply (unfold concurrent_empty_stack_def)
apply (rule conjI)
apply (assumption)
apply (unfold concurrent_empty_stack_axioms_def)
apply (rule conjI)+
apply (assumption)+
apply (rule conjI)
by (assumption)+

end

context concurrent_empty_stack
begin

lemma slin_transitive : "trans slin"
apply (insert stack_linearizable)
apply (unfold sequential_stack_def)
apply (erule  conjE)
apply (unfold sequential_set_def)
apply (erule  conjE)
apply (unfold wellformed_history_def)
apply (erule  conjE)
by metis

lemma slin_irreflexive : "irrefl slin"
apply (insert stack_linearizable)
apply (unfold sequential_stack_def)
apply (erule  conjE)
apply (unfold sequential_set_def)
apply (erule  conjE)
apply (unfold wellformed_history_def)
apply (erule  conjE)
by metis

lemma slin_antisymmetric : "antisym slin"
apply (insert stack_linearizable)
apply (unfold sequential_stack_def)
apply (erule  conjE)
apply (unfold sequential_set_def)
apply (erule  conjE)
apply (unfold wellformed_history_def)
apply (erule  conjE)

by metis

lemma slin_total : "\<not> emp a \<Longrightarrow> \<not> emp b \<Longrightarrow> a=b \<or> (a, b) \<in> slin \<or> (b, a) \<in> slin"
apply (insert stack_linearizable)
apply (unfold sequential_stack_def)
apply (erule  conjE)
apply (unfold sequential_set_def)
apply (erule  conjE)
apply (unfold sequential_set_axioms_def)
by (metis is_pop_def is_push_def lem_empty_or_push_or_pop wellformed_history.lem_val_then_push_and_pop)

lemma val_then_slin : "(a, b) \<in> val \<Longrightarrow> (a, b) \<in> slin"
apply (insert stack_linearizable)
apply (unfold sequential_stack_def)
apply (erule conjE)
apply (unfold sequential_set_def)
apply (erule conjE)
apply (unfold wellformed_history_def)
apply (erule conjE)
by (metis (hide_lams, no_types) def_empty_then_not_val empty_then_not_val_def slin_total val_order_def val_then_not_pr_def)

lemma slin_ooo : "(ia, ra) \<in> val \<Longrightarrow> (ib, rb) \<in> val \<Longrightarrow> (ia, ib) \<in> slin \<Longrightarrow> (ib, ra) \<in> slin \<Longrightarrow> (rb, ra) \<in> slin"
apply (insert stack_linearizable)
apply (unfold sequential_stack_def)
apply (erule conjE)
apply (unfold sequential_stack_axioms_def)
apply (unfold sequential_lifo_def)
by metis

lemma elin_transitive : "trans elin"
apply (insert set_linearizable)
apply (unfold sequential_empty_set_def)
apply (erule  conjE)
apply (unfold sequential_empty_history_def)
apply (erule  conjE)
by (metis wellformed_empty_history.pre_transitive)


lemma elin_irreflexive : "irrefl elin"
apply (insert set_linearizable)
apply (unfold sequential_empty_set_def)
apply (erule  conjE)
apply (unfold sequential_empty_history_def)
apply (erule  conjE)
by (metis wellformed_empty_history.pre_irreflexive)

lemma elin_antisymmetric : "antisym elin"
apply (insert set_linearizable)
apply (unfold sequential_empty_set_def)
apply (erule  conjE)
apply (unfold sequential_empty_history_def)
apply (erule  conjE)
by (metis wellformed_empty_history.pre_antisymmetric)

lemma elin_total : "a=b \<or> (a, b) \<in> elin \<or> (b, a) \<in> elin"
apply (insert set_linearizable)
apply (unfold sequential_empty_set_def)
apply (erule  conjE)
apply (unfold sequential_empty_history_def)
apply (erule  conjE)
apply (unfold sequential_empty_history_axioms_def)
by (metis UNIV_I total_on_def)

lemma elin_correct_empty : "emp a \<Longrightarrow> (ib, rb) \<in> val \<Longrightarrow> 
   ((ib, a) \<in> elin \<and> (rb, a) \<in> elin) \<or> ((a, ib) \<in> elin \<and> (a, rb) \<in> elin)"
apply (insert set_linearizable)
apply (unfold sequential_empty_set_def)
apply (erule  conjE)
apply (unfold sequential_empty_set_axioms_def)
apply (unfold correct_empty_def)
by metis

definition w1 :: "'a rel"
  where "w1 = {(a, b) . \<not> emp a \<and> \<not> emp b \<and> (\<exists> c . emp c \<and> (a, c) \<in> elin \<and> (c, b) \<in> elin)}"

lemma w1D : "(a, b) \<in> w1 \<Longrightarrow> \<not> emp a \<and> \<not> emp b \<and> (\<exists> c . emp c \<and> (a, c) \<in> elin \<and> (c, b) \<in> elin)"
apply (unfold w1_def)
by (simp add: set_eq_iff)

lemma w1I : "\<not> emp a \<Longrightarrow> \<not> emp b \<Longrightarrow> emp c \<Longrightarrow> (a, c) \<in> elin \<Longrightarrow> (c, b) \<in> elin \<Longrightarrow> (a, b) \<in> w1"
apply (unfold w1_def)
apply (simp add: set_eq_iff)
by metis

lemma w1_transitive : "trans w1"
by (metis elin_transitive trans_def w1D w1I)

lemma w1_antisymmetric : "antisym w1"
by (metis antisym_def elin_antisymmetric transE w1D w1_transitive)

lemma w1_irreflexive : "irrefl w1"
by (metis antisymD elin_antisymmetric irrefl_def w1D)

lemma w1_then_not_inv_pr : "(a, b) \<in> w1 \<Longrightarrow> \<not> (b, a) \<in> pre"
by (metis antisymD elin_antisymmetric elin_transitive pr_in_elin transD w1D)

lemma not_w1_then_same_with_empty : "(a, b) \<notin> w1 \<Longrightarrow> (b, a) \<notin> w1 \<Longrightarrow> \<not> emp a \<Longrightarrow> \<not> emp b \<Longrightarrow> emp c \<Longrightarrow> (a, c) \<in> elin \<Longrightarrow> (b, c) \<in> elin"
apply (insert elin_total)
apply (drule_tac x="b" in meta_spec)
apply (drule_tac x="c" in meta_spec)
apply (erule disjE)
apply metis
apply (erule disjE)
apply metis
by (metis w1I)

lemma val_then_not_w1 : "(a, b) \<in> val \<Longrightarrow> (a, b) \<notin> w1 \<and> (b, a) \<notin> w1"
apply (rule conjI)
apply (rule notI)
apply (drule w1D)
apply (erule conjE)+
apply (erule exE)
apply (erule conjE)+
apply (metis antisymD elin_antisymmetric elin_correct_empty)
by (metis antisym_def elin_antisymmetric elin_correct_empty w1D)

lemma val_then_same_w11 : "(ia, ra) \<in> val \<Longrightarrow> (b, ia) \<in> w1 \<Longrightarrow> (b, ra) \<in> w1"
by (smt def_empty_then_not_val elin_correct_empty empty_then_not_val_def val_then_not_w1 w1D w1I)

lemma val_then_same_w12 : "(ia, ra) \<in> val \<Longrightarrow> (b, ra) \<in> w1 \<Longrightarrow> (b, ia) \<in> w1"
by (smt def_empty_then_not_val elin_correct_empty empty_then_not_val_def val_then_not_w1 w1D w1I)

lemma val_then_same_w13 : "(ia, ra) \<in> val \<Longrightarrow> (ia, b) \<in> w1 \<Longrightarrow> (ra, b) \<in> w1"
by (smt def_empty_then_not_val elin_correct_empty empty_then_not_val_def val_then_not_w1 w1D w1I)

lemma val_then_same_w14 : "(ia, ra) \<in> val \<Longrightarrow> (ra, b) \<in> w1 \<Longrightarrow> (ia, b) \<in> w1"
by (smt def_empty_then_not_val elin_correct_empty empty_then_not_val_def val_then_not_w1 w1D w1I)

definition witness :: "'a rel"
where "witness = {(a, b) . (emp a \<and> (a, b) \<in> elin) \<or>
                           (emp b \<and> (a, b) \<in> elin) \<or>
                           (a, b) \<in> w1 \<or>
                           (\<not> emp a \<and> \<not> emp b \<and> (a, b) \<notin> w1 \<and> (b, a) \<notin> w1 \<and> (a, b) \<in> slin)}"

lemma witnessD : "(a, b) \<in> witness \<Longrightarrow> (emp a \<and> (a, b) \<in> elin) \<or>
                           (emp b \<and> (a, b) \<in> elin) \<or>
                           (a, b) \<in> w1 \<or>
                           (\<not> emp a \<and> \<not> emp b \<and> (a, b) \<notin> w1 \<and> (b, a) \<notin> w1 \<and> (a, b) \<in> slin)"
apply (unfold witness_def)
by (simp add: set_eq_iff)

lemma witnessI1 : "emp a \<Longrightarrow> (a, b) \<in> elin \<Longrightarrow> (a, b) \<in> witness"
apply (unfold witness_def)
by (simp add: set_eq_iff)

lemma witnessI2 : "emp b \<Longrightarrow> (a, b) \<in> elin \<Longrightarrow> (a, b) \<in> witness"
apply (unfold witness_def)
by (simp add: set_eq_iff)

lemma witnessI3 : "(a, b) \<in> w1 \<Longrightarrow> (a, b) \<in> witness"
apply (unfold witness_def)
by (simp add: set_eq_iff)

lemma witnessI4 : "\<not> emp a \<Longrightarrow> \<not> emp b \<Longrightarrow> (a, b) \<notin> w1 \<Longrightarrow> (b, a) \<notin> w1 \<Longrightarrow> (a, b) \<in> slin \<Longrightarrow> (a, b) \<in> witness"
apply (unfold witness_def)
by (simp add: set_eq_iff)

lemma pr_in_witness : "pre \<subseteq> witness"
apply (rule subrelI)
apply (case_tac "emp x")
apply (metis pr_in_elin witnessI1)
apply (case_tac "emp y")
apply (metis pr_in_elin witnessI2)
apply (case_tac "(x, y) \<in> w1")
apply (rule witnessI3)
apply (assumption)
by (metis pr_in_slin w1_then_not_inv_pr witnessI4)

lemma witness_transitive : "trans witness"
apply (rule transI)
apply (drule witnessD)+
apply (erule disjE)+
apply (metis elin_transitive transE witnessI1)
apply (erule disjE)
apply (metis elin_transitive transD witnessI1)
apply (erule disjE)
apply (metis elin_transitive trans_def w1D witnessI1)
apply (erule conjE)+
apply (metis elin_total w1I witnessI1)
apply (erule disjE)+
apply (metis elin_transitive transE w1I witnessI1 witnessI2 witnessI3)
apply (erule disjE)+
apply (metis w1D)
apply metis
apply (erule disjE)
apply (metis elin_transitive trans_def w1D witnessI2)
apply (erule disjE)
apply (metis antisym_def elin_total w1D w1I w1_antisymmetric witnessI2)
apply (erule disjE)+
apply (metis transD w1_transitive witnessI3)
apply (metis not_w1_then_same_with_empty transE w1D w1I w1_transitive witnessI3)
apply (erule disjE)
apply (metis not_w1_then_same_with_empty w1D w1I witnessI3)
by (metis not_w1_then_same_with_empty slin_transitive transD w1D w1I witnessI4)

lemma witness_irreflexive : "irrefl witness"
by (metis elin_irreflexive irrefl_def slin_irreflexive w1_irreflexive witnessD)

lemma witness_antisymmetric : "antisym witness"
by (metis antisymI lem_irreflexive_then_not_equal trans_def witness_irreflexive witness_transitive)

lemma witness_total : "total witness"
by (metis elin_total slin_total total_on_def witnessI1 witnessI2 witnessI3 witnessI4)

lemma witness_correct_empty : "emp a \<Longrightarrow> (ib, rb) \<in> val \<Longrightarrow> 
   ((ib, a) \<in> witness \<and> (rb, a) \<in> witness) \<or> ((a, ib) \<in> witness \<and> (a, rb) \<in> witness)"
by (metis elin_correct_empty witnessI1 witnessI2)

lemma val_then_not_witness : "(ia, ra) \<in> val \<Longrightarrow> (ra, ia) \<notin> witness"
by (metis antisym_def def_empty_then_not_val def_val_order empty_then_not_val_def slin_antisymmetric val_order_def val_then_not_w1 val_then_slin witnessD)

lemma witness_ooo : "(ia, ra) \<in> val \<Longrightarrow> (ib, rb) \<in> val \<Longrightarrow> (ia, ib) \<in> witness \<Longrightarrow> (ib, ra) \<in> witness \<Longrightarrow> (rb, ra) \<in> witness"
apply (drule witnessD)+
apply (erule disjE)+
apply (metis def_empty_then_not_val empty_then_not_val_def)
apply (erule disjE)
apply (metis def_empty_then_not_val empty_then_not_val_def)
apply (erule disjE)
apply (metis val_then_same_w12 w1D)
apply (metis elin_correct_empty elin_irreflexive irrefl_def)
apply (erule disjE)+
apply (metis def_empty_then_not_val empty_then_not_val_def)
apply (erule disjE)
apply (metis w1D)
apply metis
apply (erule disjE)+
apply (metis def_empty_then_not_val empty_then_not_val_def)
apply (erule disjE)
apply (metis w1D)
apply metis
apply (erule disjE)+
apply (metis val_then_same_w13 w1D)
apply (metis elin_correct_empty elin_irreflexive irrefl_def)
apply (erule disjE)+
apply (metis val_then_same_w13 witnessI3)
apply (metis val_then_same_w13)
apply (erule disjE)
apply (metis val_then_same_w12)
by (metis def_empty_then_not_val empty_then_not_val_def slin_ooo val_then_same_w12 val_then_same_w14 witnessI4)

lemma witness_sequential_empty_stack : "sequential_empty_stack witness val emp f"
apply (rule sequential_empty_stackI)
apply (rule sequential_empty_historyI)
apply (rule wellformed_empty_historyI)
apply (metis witness_transitive)
apply (metis witness_antisymmetric)
apply (metis witness_irreflexive)
apply (metis operations_countable)
apply (metis operations_finite)
apply (metis def_push_once)
apply (metis def_pop_once)
apply (metis def_val_order)
apply (metis def_all_popped_or_empty)
apply (metis def_empty_then_not_val)
apply (metis val_then_not_prI val_then_not_witness)
apply (metis witness_total)
apply (smt sequential_lifo_def witness_ooo)
by (smt correct_empty_def witness_correct_empty)

theorem concurrent_empty_stack_linearizable : "\<exists> prt .( pre \<subseteq> prt \<and> sequential_empty_stack prt val emp f)"
apply (rule_tac x="witness" in exI)
apply (rule conjI)
apply (rule pr_in_witness)
by (rule witness_sequential_empty_stack)

end

lemma stack_set_then_set: "linearizable_non_empty_stack (non_empty_pre pre emp) val \<Longrightarrow> linearizable_set pre val emp \<Longrightarrow>
  \<exists> elin slin f . concurrent_empty_stack pre val emp f slin elin"
apply (unfold linearizable_non_empty_stack_def)
apply (erule conjE)+
apply (erule exE)+
apply (rename_tac slin f)
apply (erule conjE)
apply (unfold linearizable_set_def)
apply (erule conjE)+
apply (erule exE)+
apply (rename_tac elin fa)
apply (erule conjE)
apply (rule_tac x="elin" in exI)
apply (rule_tac x="slin" in exI)
apply (rule_tac x="f" in exI)
apply (frule sequential_empty_setD)
apply (erule conjE)
apply (frule sequential_empty_historyD)
apply (erule conjE)
apply (frule wellformed_empty_historyD)
apply (erule conjE)+
apply (rule wellformed_empty_history.concurrent_empty_stackI)
apply (rule wellformed_empty_historyI)
apply (metis conc_historyD)
apply (metis conc_historyD)
apply (metis conc_historyD)
apply (metis sequential_setD sequential_stackD wellformed_historyD)
apply (assumption)+
apply (metis set_rev_mp val_then_not_prI val_then_not_pr_def)
apply (rule wellformed_empty_historyI)
apply (metis conc_historyD)
apply (metis conc_historyD)
apply (metis conc_historyD)
apply (metis sequential_setD sequential_stackD wellformed_historyD)
apply (assumption)+
apply (metis in_mono val_then_not_pr_def)
apply (assumption)+
apply (rule allI)+
apply (rule impI)+
apply (frule_tac a="a" and b="b" and pre="pre" and emp="emp" in non_empty_preI)
apply (assumption)+
apply (metis in_mono)
apply (rule sequential_empty_setI)
apply (rule sequential_empty_historyI)
apply (rule wellformed_empty_historyI)
apply (drule wellformed_empty_historyD)
apply (drule sequential_empty_setD)
apply (drule sequential_empty_historyD)
apply (erule conjE)+
apply (assumption)+
apply (metis sequential_setD sequential_stackD wellformed_historyD)
apply (assumption)+
by (metis set_rev_mp)

theorem concurrent_empty_stack_linearizable : "linearizable_non_empty_stack (non_empty_pre pre emp) val \<Longrightarrow> linearizable_set pre val emp
 \<Longrightarrow> linearizable_stack pre val emp"
apply (drule stack_set_then_set)
apply (assumption)
apply (erule exE)
apply (erule exE)
by (metis concurrent_empty_stack.concurrent_empty_stack_linearizable linearizable_set_def linearizable_stack_def)

end
