theory insert_pr_relaxed_stack
imports Main insert_relaxed_stack
begin

locale insert_pr_relaxed_stack = wellformed_history +
  fixes ir :: "'a rel"
  assumes val_then_pr: "(a, b) \<in> val \<Longrightarrow> (a, b) \<in> pre"
  assumes pop_ordered: "is_pop a \<Longrightarrow> is_pop b \<Longrightarrow>a\<noteq>b\<Longrightarrow>(a, b) \<in> pre \<or> (b,a)\<in>pre"
  assumes ir_alternating: "alternating val ir"
  assumes pr_u_ir_acyclic: "acyclic (pre \<union> ir)"
  assumes ipr_ooo : "(ia, ra) \<in> val \<Longrightarrow> (ib, rb) \<in> val \<Longrightarrow> (ia, ib) \<in> (ins_order pre val ir) \<Longrightarrow> (ib, ra) \<in> ir \<Longrightarrow> (rb, ra) \<in> pre"

context insert_pr_relaxed_stack
begin

lemma push_pop_ordered : "is_push a \<Longrightarrow>is_pop b \<Longrightarrow> (a, b) : ir \<or> (b,a) \<in> ir"
by (metis alternating_def ir_alternating isPop_def isPush_def is_popD is_pushD)

lemma ir_push_and_pop : "(a, b) \<in> ir \<Longrightarrow> (is_push a \<and> is_pop b) \<or> (is_pop a \<and> is_push b)"
by (metis alternating_def ir_alternating isPop_def isPush_def is_pop_def is_push_def)

definition ins :: "'a rel"
where
  "ins = ins_order pre val ir"

lemma lem_pr_then_ins : " is_push a \<Longrightarrow> is_push b \<Longrightarrow> (a, b) \<in> pre \<Longrightarrow> (a, b) \<in> ins"
apply (unfold ins_def)
apply (unfold ins_order_def)
apply (simp add: set_eq_iff)
by (metis isPush_def is_pushD)

lemma insI : "is_push a \<Longrightarrow> is_push b \<Longrightarrow> (a, c) \<in> pre \<Longrightarrow> (c, b) \<in> ir \<Longrightarrow> (a, b) \<in> ins"
by (metis ins_def ins_orderI2 isPush_def is_pushD)

lemma ir_push_and_pop1 : "(a, b) \<in> ir \<Longrightarrow> is_push a \<Longrightarrow> is_pop b"
by (metis def_value ir_push_and_pop is_push_def lem_pop_then_push_exists val_order_def)

lemma ir_push_and_pop2 : "(a, b) \<in> ir \<Longrightarrow> is_push b \<Longrightarrow> is_pop a"
by (metis def_value ir_push_and_pop is_push_def lem_pop_then_push_exists val_order_def)

lemma ir_push_and_pop3 : "(a, b) \<in> ir \<Longrightarrow> is_pop a \<Longrightarrow> is_push b"
by (metis def_value ir_push_and_pop is_push_def lem_pop_then_push_exists val_order_def)

lemma ir_push_and_pop4 : "(a, b) \<in> ir \<Longrightarrow> is_pop b \<Longrightarrow> is_push a"
by (metis def_value ir_push_and_pop is_push_def lem_pop_then_push_exists val_order_def)

lemma lem_val_then_ir : "(a, b) \<in> val \<Longrightarrow> (a, b) \<in> ir"
apply (frule val_then_pr)
apply (frule lem_val_then_push_and_pop)
apply (erule conjE)
apply (frule_tac a="a" and b="b" in push_pop_ordered)
apply (assumption)
apply (erule disjE)
apply (assumption)
apply (frule_tac c="(a, b)" and A="pre" and B="ir" in UnI1)
apply (frule_tac c="(b, a)" and A="pre" and B="ir" in UnI2)
apply (frule_tac a="a" and b="b" and c="a" and R ="pre Un ir" in r_r_into_trancl)
apply (assumption)+
by (metis acyclic_def pr_u_ir_acyclic)

lemma lem_push_pr_pop_then_ir : " is_push a \<Longrightarrow> is_pop b \<Longrightarrow> (a, b) \<in> pre \<Longrightarrow> (a, b) \<in> ir"
apply (frule_tac a="a" and b="b" in push_pop_ordered)
apply (assumption)
apply (erule disjE)
apply (assumption)
apply (frule_tac c="(a, b)" and A="pre" and B="ir" in UnI1)
apply (frule_tac c="(b, a)" and A="pre" and B="ir" in UnI2)
apply (frule_tac a="a" and b="b" and c="a" and R ="pre Un ir" in r_r_into_trancl)
apply (assumption)+
by (metis acyclic_def pr_u_ir_acyclic)

lemma not_ir_push_pop_ir : "(a, b) \<notin> ir \<Longrightarrow> is_push a \<Longrightarrow> is_pop b \<Longrightarrow> (b, a) \<in> ir"
by (metis push_pop_ordered)

lemma not_ir_pop_push_ir : "(a, b) \<notin> ir \<Longrightarrow> is_pop a \<Longrightarrow> is_push b \<Longrightarrow> (b, a) \<in> ir"
by (metis push_pop_ordered)

lemma lem_pop_pr_push_then_ir : " is_pop a \<Longrightarrow> is_push b \<Longrightarrow> (a, b) \<in> pre \<Longrightarrow> (a, b) \<in> ir"
apply (frule_tac a="b" and b="a" in push_pop_ordered)
apply (assumption)
apply (erule disjE)
apply (frule_tac c="(a, b)" and A="pre" and B="ir" in UnI1)
apply (frule_tac c="(b, a)" and A="pre" and B="ir" in UnI2)
apply (frule_tac a="a" and b="b" and c="a" and R ="pre Un ir" in r_r_into_trancl)
apply (assumption)+
by (metis acyclic_def pr_u_ir_acyclic)

inductive_set wir :: "'a rel"
where
  wir_intro   : "(a, b) \<in> pre \<Longrightarrow> is_push a \<Longrightarrow> is_pop b \<Longrightarrow> (a, b) \<in> wir"        
| wir_step: "(\<exists>ix rx ra. (ix, rx) \<in> val \<and>
                          (a, ra) \<in> val \<and>
                          is_pop b \<and>
                          (ix, b) \<in> wir \<and>
                          (rx, ra) \<in> pre \<and>
                          (b, rx) \<in> pre \<and>
                          (a, rx) \<in> wir) \<Longrightarrow> (a, b) \<in> wir"


lemma lem_wir_pr_then_wir_help : "(a, b) \<in> wir \<Longrightarrow>
    is_push c \<or> (b, c) \<notin> pre \<or>
   (is_pop c \<and> (b, c) \<in> pre \<and> (a, c) \<in> wir)"
apply (insert pre_transitive)
apply (erule wir.induct)
apply (metis lem_push_or_pop2 transE wir.wir_intro)
apply (erule exE)+
apply (erule conjE)+
apply (erule disjE)
apply (simp)
apply (erule disjE)
apply metis
by (metis lem_val_then_push_and_pop pop_ordered wir.wir_step)

lemma lem_wir_pr_then_wir : "(a, b) \<in> wir \<Longrightarrow> is_pop c \<Longrightarrow> (b, c) \<in> pre \<Longrightarrow> (a, c) \<in> wir"
by (metis def_value lem_wir_pr_then_wir_help lem_pop_then_push_exists lem_push_then_pop_exists val_order_def)


lemma wir_right_rec : "(ix, rx) \<in> val \<Longrightarrow> (a, ra) \<in> val \<Longrightarrow> is_pop b \<Longrightarrow> 
             (ix, b) \<in> pre \<Longrightarrow> (rx, ra) \<in> pre \<Longrightarrow> (a, rx) \<in> wir \<Longrightarrow> (a, b) \<in> wir"
by (smt lem_val_then_push_and_pop lem_wir_pr_then_wir pop_ordered wir.wir_intro wir.wir_step)

lemma wir_left_rec : "(ix, rx) \<in> val \<Longrightarrow> (a, ra) \<in> val \<Longrightarrow> is_pop b \<Longrightarrow> 
                      (ix, b) \<in> wir \<Longrightarrow> (rx, ra) \<in> pre \<Longrightarrow> (a, rx) \<in> pre \<Longrightarrow> (a, b) \<in> wir"
by (smt def_pop_once ir_push_and_pop1 is_push_def lem_val_then_ir lem_val_then_push_and_pop lem_wir_pr_then_wir pop_once_def pop_ordered val_then_pr wir.simps wir.wir_intro wir.wir_step wir_right_rec)

lemma wir_left_right_rec : "(ix, rx) \<in> val \<Longrightarrow> (a, ra) \<in> val \<Longrightarrow> is_pop b \<Longrightarrow> (ix, b) \<in> wir \<Longrightarrow> (rx, ra) \<in> pre \<Longrightarrow> (a, rx) \<in> wir \<Longrightarrow> (a, b) \<in> wir"
by (metis (lifting, no_types) lem_val_then_push_and_pop lem_wir_pr_then_wir pop_ordered wir.wir_step)


lemma wirD : "(ia, rb) \<in> wir \<Longrightarrow> (ia, rb) \<in> pre \<or> 
                  (\<exists>ix rx ra. (ix, rx) \<in> val \<and>
                          (ia, ra) \<in> val \<and>
                          is_pop rb \<and>
                          (ix, rb) \<in> wir \<and>
                          (rx, ra) \<in> pre \<and>
                          (rb, rx) \<in> pre \<and>
                          (ia, rx) \<in> wir)"
by (metis wir.simps)

inductive_set lwir :: "'a rel"
where
  lwir_intro   : "(a, b) \<in> pre \<Longrightarrow> is_push a \<Longrightarrow> is_pop b \<Longrightarrow> (a, b) \<in> lwir"
| lwir_step: "(\<exists>ix rx ra. (ix, rx) \<in> val \<and>
                          (a, ra) \<in> val \<and>
                          is_pop b \<and>
                          (ix, b) \<in> lwir \<and>
                          (rx, ra) \<in> pre \<and>
                          (b, rx) \<in> pre \<and>
                          (a, rx) \<in> pre) \<Longrightarrow> (a, b) \<in> lwir"

lemma lem_lwir_then_wir : "(a, b) \<in> lwir \<Longrightarrow> (a, b) \<in> wir"
apply (erule lwir.induct)
apply (metis wir.wir_intro)
by (metis wir_left_rec)

lemma lem_lwir_left_grow : "(ib, rc) \<in> lwir \<Longrightarrow> (\<exists> rb ic.
                            (ia, ra) \<in> val \<and> (ib, rb) \<in> val \<and> (ic, rc) \<in> val \<and> 
                            (rc, rb) \<in> pre \<and> (rb, ra) \<in> pre \<and>
                            (ia, rb) \<in> pre) \<longrightarrow>  (ia, rc) \<in> lwir"
by (metis lem_val_then_push_and_pop lwir.lwir_step)

lemma lem_lwir_trans : "(ia, rb) \<in> lwir \<Longrightarrow> (\<exists> ra ib . (ic, rc) \<in> val \<and>
   (ia, ra) \<in> val \<and> (ib, rb) \<in> val \<and> (ib, rc) \<in> lwir \<and> (rc, rb) \<in> pre \<and>
   (rb, ra) \<in> pre) \<longrightarrow> (ia, rc) \<in> lwir"
apply (erule lwir.induct)
apply (smt lem_lwir_left_grow)
by (smt pre_transitive def_pop_once lem_lwir_left_grow pop_once_def transE)

lemma lem_wir_then_lwir : "(a, b) \<in> wir \<Longrightarrow> (a, b) \<in> lwir"
apply (erule wir.induct)
apply (metis lwir.lwir_intro)
by (metis lem_pop_then_push_exists lem_lwir_trans)

inductive_set rwir :: "'a rel"
where
  rwir_intro   : "(a, b) \<in> pre \<Longrightarrow> is_push a \<Longrightarrow> is_pop b \<Longrightarrow> (a, b) \<in> rwir"
| rwir_step: "(\<exists>ix rx ra. (ix, rx) \<in> val \<and>
                          (a, ra) \<in> val \<and>
                          is_pop b \<and>
                          (ix, b) \<in> pre \<and>
                          (rx, ra) \<in> pre \<and>
                          (b, rx) \<in> pre \<and>
                          (a, rx) \<in> rwir) \<Longrightarrow> (a, b) \<in> rwir"

lemma lem_rwir_then_wir : "(a, b)\<in>rwir \<Longrightarrow> (a, b) \<in> wir"
apply (erule rwir.induct)
apply (metis wir.wir_intro)
by (metis wir_right_rec)



lemma lem_left_grow : "(ia, ra) \<in> val \<Longrightarrow> (ib, rb) \<in> val \<Longrightarrow> (ic, rc) \<in> val \<Longrightarrow> 
                       (ia, rb) \<in> wir \<Longrightarrow> (ic, ra) \<in> pre \<Longrightarrow> (ra, rc) \<in> pre \<Longrightarrow> (ic, rb) \<in> wir"
by (metis lem_val_then_push_and_pop wir_left_rec)

lemma lem_right_grow : "(ia, ra) \<in> val \<Longrightarrow> (ib, rb) \<in> val \<Longrightarrow> (ic, rc) \<in> val \<Longrightarrow> (rb, ra) \<in> pre \<Longrightarrow>
                       (ia, rb) \<in> wir \<Longrightarrow> (ib, rc) \<in> pre \<Longrightarrow> (rc, rb) \<in> pre \<Longrightarrow> (ia, rc) \<in> wir"
by (metis lem_val_then_push_and_pop wir_right_rec)

lemma lem_rwir_right_grow : "(ia, ra) \<in> val \<Longrightarrow> (ib, rb) \<in> val \<Longrightarrow> (ic, rc) \<in> val \<Longrightarrow> (rc, rb) \<in> pre \<Longrightarrow>
                       (ia, rb) \<in> rwir \<Longrightarrow> (ib, rc) \<in> pre \<Longrightarrow> (rb, ra) \<in> pre \<Longrightarrow> (ia, rc) \<in> rwir"
by (metis lem_val_then_push_and_pop rwir.rwir_step)

lemma lem_rwir_left_grow : "(ib, rc) \<in> rwir \<Longrightarrow> (\<exists> rb ic.
                            (ia, ra) \<in> val \<and> (ib, rb) \<in> val \<and> (ic, rc) \<in> val \<and> 
                            (rc, rb) \<in> pre \<and> (rb, ra) \<in> pre \<and>
                            (ia, rb) \<in> pre) \<longrightarrow>  (ia, rc) \<in> rwir"
apply (erule rwir.induct)
apply (rename_tac ib rc)
apply (rule impI)
apply (erule exE)+
apply (erule conjE)+
apply (drule_tac a="ia" and b="rb" in rwir.rwir_intro)
apply (metis lem_val_then_push_and_pop)
apply (metis is_pop_def)
apply (metis rwir.rwir_step)
apply (rename_tac ib rc)
apply (safe)
apply (rename_tac rb rbb ic)
apply (metis def_pop_once pop_once_def)
by (smt pre_transitive def_pop_once lem_rwir_right_grow pop_once_def transE)

lemma lem_rwir : "(ia, rb) \<in> rwir \<Longrightarrow> (ia, rb) \<in> pre \<or> 
                  (\<exists>ix rx ra. (ix, rx) \<in> val \<and>
                          (ia, ra) \<in> val \<and>
                          is_pop rb \<and>
                          (ix, rb) \<in> pre \<and>
                          (rx, ra) \<in> pre \<and>
                          (rb, rx) \<in> pre \<and>
                          (ia, rx) \<in> rwir)"
by (metis rwir.simps)

lemma lem_rwir_trans : "(ib, rc) \<in> rwir \<Longrightarrow> (\<exists> rb ic . (ia, ra) \<in> val \<and>
   (ib, rb) \<in> val \<and> (ic, rc) \<in> val \<and> (ia, rb) \<in> rwir \<and> (rb, ra) \<in> pre \<and>
   (rc, rb) \<in> pre) \<longrightarrow> (ia, rc) \<in> rwir"
apply (erule rwir.induct)
apply (rename_tac ib rc)
apply (smt lem_rwir_right_grow)
by (smt pre_transitive def_pop_once lem_rwir_right_grow pop_once_def transE)

lemma lem_wir_then_rwir : "(a, b) \<in> wir \<Longrightarrow> (a, b) \<in> rwir"
apply (erule wir.induct)
apply (metis rwir.rwir_intro)
by (metis lem_pop_then_push_exists lem_rwir_trans)

lemma lem_rwir_then_push_and_pop : "(a, b) \<in> rwir \<Longrightarrow> is_push a \<and> is_pop b"
by (metis is_push_def rwir.simps)


definition wri :: "'a rel"
where
  "wri = {(a, b). is_pop a \<and> is_push b \<and> (b, a) \<notin> wir}"

definition wii :: "'a rel"
where 
  " wii = {(a, b). is_push a \<and> is_push b \<and> (a, b) \<in> pre}"

definition wrr :: "'a rel"
where
  "wrr = {(a, b). is_pop a \<and> is_pop b \<and> (a, b) \<in> pre}"

definition witness :: "'a rel"
where
  "witness = trancl (wir \<union> wri \<union> wii \<union> wrr)"

lemma lem_wir_then_push_and_pop : "(a, b) \<in> wir \<Longrightarrow> is_push a \<and> is_pop b"
by (metis is_push_def wir.simps)

lemma lem_wri_then_pop_and_push : "(a, b) \<in> wri \<Longrightarrow> is_pop a \<and> is_push b"
by (metis (lifting) internal_split_conv internal_split_def mem_Collect_eq wri_def)

lemma lem_wir_wri_total : "is_push a \<Longrightarrow> is_pop b \<Longrightarrow> (a, b) \<in> wir \<or> (b, a) \<in> wri"
apply (case_tac "(a, b) \<in> wir")
apply (rule disjI1)
apply (assumption)
apply (rule disjI2)
apply (unfold wri_def)
by (auto)

lemma lem_wrr_total : "is_pop a \<Longrightarrow> is_pop b \<Longrightarrow> a \<noteq> b \<Longrightarrow> (a, b) \<in> wrr \<or> (b, a) \<in> wrr"
apply (frule_tac a="a" and b="b" in pop_ordered)
apply (assumption)+
apply (erule disjE)
apply (rule disjI1)
apply (unfold wrr_def)
by (safe)

lemma lem_wir_then_not_wri : "is_push a \<Longrightarrow> is_pop b \<Longrightarrow> (a, b) \<in> wir \<Longrightarrow> (b, a) \<notin> wri"
apply (rule notI)
apply (unfold wri_def)
by (safe)

lemma wriI : "is_push a \<Longrightarrow> is_pop b \<Longrightarrow> (a, b) \<notin> wir \<Longrightarrow> (b, a) \<in> wri"
by (metis lem_wir_wri_total)

lemma wriD : "(a, b) \<in> wri \<Longrightarrow> is_pop a \<and> is_push b \<and> (b, a) \<notin> wir"
by (metis lem_wir_then_not_wri lem_wri_then_pop_and_push)

lemma wrrD : "(a, b) \<in> wrr \<Longrightarrow> is_pop a \<and> is_pop b \<and> (a, b)\<in> pre"
by (metis (lifting) mem_Collect_eq split_conv wrr_def)

lemma wiiD : "(a, b) \<in> wii \<Longrightarrow> is_push a \<and> is_push b \<and> (a, b)\<in> pre"
apply (unfold wii_def)
by (simp add: set_eq_iff)

lemma lem_wri : "(rb, ia) \<in> wri \<Longrightarrow> (ia, ra) \<in> val \<Longrightarrow> (ix, rx) \<in> val \<Longrightarrow> 
  \<not>((ia, rb) \<in> pre \<or> ((ix, rb) \<in> pre \<and> (rx, ra) \<in> pre \<and> (ia, rx) \<in> wir))"
by (metis is_push_def lem_val_then_push_and_pop wir.wir_intro wir_right_rec wriD)

lemma lem_rwir_then_ir : "(a, b) \<in> rwir \<Longrightarrow> (a, b) \<in> ir"
apply (erule rwir.induct)
apply (metis lem_push_pr_pop_then_ir)
apply (erule exE)+
apply (erule conjE)+
apply (rotate_tac -3)
apply (erule contrapos_pp)
apply (frule_tac a="ix" and b="rx" in lem_val_then_push_and_pop)
apply (frule_tac a="a" and b="ra" in lem_val_then_push_and_pop)
apply (erule conjE)+
apply (frule_tac a="ix" and b="a" and c="b" in insI)
apply (assumption)+
apply (metis push_pop_ordered)
by (metis pre_antisymmetric pre_irreflexive ins_def antisymD ipr_ooo lem_irreflexive_then_not_equal)

lemma lem_wir_then_ir : "(a, b) \<in> wir \<Longrightarrow> (a, b) \<in> ir"
by (metis lem_rwir_then_ir lem_wir_then_rwir)

lemma lem_ir_then_wri : "(a, b) \<in> ir \<Longrightarrow> is_pop a \<Longrightarrow> is_push b \<Longrightarrow> (a, b)\<in> wri"
apply (unfold wri_def)
apply (simp add: set_eq_iff)
apply (erule contrapos_pn)
apply (drule lem_wir_then_ir)
apply (rule notI)
apply (frule_tac c="(a, b)" and A="pre" and B="ir" in UnI2)
apply (frule_tac c="(b, a)" and A="pre" and B="ir" in UnI2)
apply (frule_tac a="a" and b="b" and c="a" and R ="pre Un ir" in r_r_into_trancl)
apply (assumption)+
by (metis acyclic_def pr_u_ir_acyclic)

lemma lem_wir_then_not_rev_pr : "(a, b) \<in> wir \<Longrightarrow> (b, a) \<notin> pre"
apply (erule wir.induct)
apply (metis pre_antisymmetric pre_irreflexive antisymD lem_irreflexive_then_not_equal)
by (metis lem_ir_then_wri lem_pop_pr_push_then_ir lem_val_then_push_and_pop wir_left_right_rec wriD)

lemma lem_pr_then_wri : "(a, b) \<in> pre \<Longrightarrow> is_pop a \<Longrightarrow> is_push b \<Longrightarrow> (a, b) \<in> wri"
apply (rule wriI)
apply (assumption)+
apply (erule contrapos_pn)
by (metis lem_wir_then_not_rev_pr)

lemma lem_pr_in_witness: "pre \<subseteq> witness"
apply (rule subrelI)
apply (unfold witness_def)
apply (rule r_into_trancl)
apply (simp)
apply (frule lem_push_or_pop1)
apply (frule lem_push_or_pop2)
apply (erule disjE)+
-- {* Subgoal 1: is_push x and is_push y *}
apply (rule disjI2)
apply (rule disjI2)
apply (rule disjI1)
apply (unfold wii_def)[1]
apply (simp)
-- {* Subgoal 2: is_push x and is_pop y *}
apply (drule_tac a="x" and b="y" in wir.wir_intro)
apply (assumption)+
apply (rule disjI1)
apply (assumption)
apply (erule disjE)
-- {* Subgoal 3: is_pop x and is_push y *}
apply (drule_tac a="x" and b="y" in lem_pr_then_wri)
apply (assumption)+
apply (rule disjI2)
apply (rule disjI1)
apply (assumption)
-- {* Subgoal 4: is_pop x and is_pop y *}
apply (rule disjI2)+
apply (insert pre_irreflexive)
apply (drule_tac r="pre" and a="x" and b="y" in lem_irreflexive_then_not_equal)
apply (assumption)
apply (unfold wrr_def)
by (safe)

lemma lem_wri_wir_then_pr : "(ra, ib) \<in> wri \<Longrightarrow> (ib, rc) \<in> wir \<Longrightarrow> (ra, rc) \<in> pre"
apply (frule lem_wri_then_pop_and_push)
apply (frule lem_wir_then_push_and_pop)
apply (erule conjE)+
apply (case_tac "ra = rc")
apply (metis lem_wir_then_not_wri)
apply (frule_tac a="ra" and b="rc" in pop_ordered)
apply (assumption)
apply (assumption)
by (metis lem_wir_pr_then_wir lem_wir_then_not_wri)

lemma lem_pr_wri_then_wri : "(ra, rb) \<in> wrr \<Longrightarrow> (rb, ic) \<in> wri \<Longrightarrow> (ra, ic) \<in> wri"
apply (rule wriI)
apply (metis lem_wri_then_pop_and_push)
apply (metis wrrD)
apply (drule wriD)
apply (erule conjE)+
apply (rotate_tac -1)
apply (erule contrapos_nn)
by (metis lem_wir_pr_then_wir wrrD)

lemma lem_ir_wir_then_pr : "(a, b) \<in> rwir \<Longrightarrow> (c, a) \<in> ir \<longrightarrow> (c, b) \<in> pre"
by (metis ir_push_and_pop2 lem_ir_then_wri lem_rwir_then_wir lem_wir_then_push_and_pop lem_wri_wir_then_pr)

lemma lem_ooo_infer_ir : "(ia, ra) \<in> val \<Longrightarrow> (ib, rb) \<in> val \<Longrightarrow> (ia, ib) \<in> ins \<Longrightarrow> (ra, rb) \<in> pre \<Longrightarrow> (ra, ib) \<in> ir"
apply (rotate_tac -1)
apply (erule contrapos_pp)
apply (drule not_ir_pop_push_ir)
apply (metis is_pop_def)
apply (metis is_push_def)
apply (drule_tac ia="ia" and ra="ra" and ib="ib" and rb="rb" in ipr_ooo)
apply (assumption)
apply (metis ins_def)
apply (assumption)
by (metis pre_antisymmetric antisymD pre_irreflexive lem_irreflexive_then_not_equal)

lemma lem_wir_then_pop_pr : "(a, b) \<in> wir \<Longrightarrow> (a, ra) \<in> val \<Longrightarrow> (a, b) \<notin> pre \<Longrightarrow> (b, ra) \<in> pre"
apply (drule wirD)
apply (erule disjE)
apply metis
apply (erule exE)+
apply (erule conjE)+
by (metis pre_transitive def_pop_once pop_once_def transD)

lemma lem_pr_wir_then_wir : "(a, b) \<in> wir \<Longrightarrow> is_push c \<Longrightarrow> (c, a) \<in> pre \<Longrightarrow> (c, b) \<in> wir"
apply (frule lem_wir_then_push_and_pop)
apply (erule conjE)
apply (frule lem_push_then_pop_exists)
apply (erule exE)
apply (rename_tac rc)
apply (erule conjE)
apply (frule_tac a="a" in lem_push_then_pop_exists)
apply (erule exE)
apply (rename_tac ra)
apply (erule conjE)
apply (frule_tac a="ra" and b="rc" in pop_ordered)
apply (assumption)
apply (metis ins_def ipr_ooo lem_ir_then_wri lem_ooo_infer_ir lem_pr_then_ins lem_val_then_ir val_then_pr wir.wir_intro wriD)
apply (erule disjE)
apply (case_tac "(a, b) \<in> pre")
apply (metis pre_transitive transE wir.wir_intro)
apply (rule wir.wir_step)
apply (rule_tac x="a" in exI)
apply (rule_tac x="ra" in exI)
apply (rule_tac x="rc" in exI)
apply (rule conjI)
apply (assumption)
apply (rule conjI)
apply (assumption)
apply (rule conjI)
apply (assumption)
apply (rule conjI)
apply (assumption)
apply (rule conjI)
apply (assumption)
apply (metis pre_transitive lem_wir_then_pop_pr transE val_then_pr wir.wir_intro)
by (metis lem_ir_then_wri lem_ooo_infer_ir lem_pr_then_ins lem_wir_pr_then_wir lem_wri_wir_then_pr val_then_pr wir.wir_intro)

lemma lem_wri_pr_then_wri : "(ra, ib)\<in> wri \<Longrightarrow> (ib, ic)\<in> wii \<Longrightarrow> (ra, ic)\<in> wri"
apply (rule wriI)
apply (metis wiiD)
apply (metis lem_wri_then_pop_and_push)
by (metis lem_pr_wir_then_wir wiiD wriD)

lemma lem_wir_wri_then_not_pr : "(ia, rb) \<in> wir \<Longrightarrow> (rb, ic) \<in> wri \<Longrightarrow> (ic, ia) \<notin> pre"
by (metis lem_pr_wir_then_wir lem_wri_then_pop_and_push wriD)

lemma witnessD : "(a, b) \<in> witness \<Longrightarrow>
  (a, b) \<in> pre \<or>
  (is_push a \<and> is_push b \<and> (\<exists>c. (a, c) \<in> wir \<and> (c, b) \<in> wri)) \<or>
  (is_push a \<and> is_pop b \<and> (a, b) \<in> wir) \<or>
  (is_pop a \<and> is_push b \<and> (a, b) \<in> wri)"
apply (unfold witness_def)
apply (erule trancl.induct)
apply (metis UnE lem_wir_then_push_and_pop wiiD wriD wrrD)
apply (erule disjE)
apply (erule UnE)+
apply (metis lem_pr_then_wri lem_pr_wir_then_wir lem_push_or_pop1 lem_wir_then_push_and_pop lem_wri_wir_then_pr)
apply (metis lem_push_or_pop1 lem_wir_pr_then_wir wir.wir_intro wriD wriI)
apply (metis pre_transitive transE wiiD)
apply (metis pre_transitive transE wrrD)
apply (erule UnE)+
apply (metis def_value is_pop_def is_push_def lem_wir_pr_then_wir_help lem_wir_then_push_and_pop lem_wri_wir_then_pr val_order_def)
apply (metis def_value is_pop_def lem_push_then_pop_exists lem_wri_then_pop_and_push val_order_def)
apply (metis lem_wir_then_not_rev_pr lem_wri_pr_then_wri wiiD wriI)
by (metis def_value lem_pop_then_push_exists lem_push_then_pop_exists lem_wir_pr_then_wir_help val_order_def wrrD)

lemma witnessI1 : "(a, b) \<in> pre \<Longrightarrow> (a, b) \<in> witness"
by (metis lem_pr_in_witness set_rev_mp)

lemma witnessI2 : "(a, c) \<in> wir \<Longrightarrow> (c, b) \<in> wri \<Longrightarrow> (a, b) \<in> witness"
thm witness_def
apply (drule_tac c="(a, c)" and A="wir" and B="wri" in UnI1)
apply (drule_tac c="(a, c)" and A="wir \<union> wri" and B="wii" in UnI1)
apply (drule_tac c="(a, c)" and A="wir \<union> wri \<union> wii" and B="wrr" in UnI1)
apply (drule_tac c="(c, b)" and B="wri" and A="wir" in UnI2)
apply (drule_tac c="(c, b)" and A="wir \<union> wri" and B="wii" in UnI1)
apply (drule_tac c="(c, b)" and A="wir \<union> wri \<union> wii" and B="wrr" in UnI1)
by (metis r_r_into_trancl witness_def)

lemma witnessI3 : "(a, b) \<in> wir \<Longrightarrow> (a, b) \<in> witness"
apply (drule_tac c="(a, b)" and A="wir" and B="wri" in UnI1)
apply (drule_tac c="(a, b)" and A="wir \<union> wri" and B="wii" in UnI1)
apply (drule_tac c="(a, b)" and A="wir \<union> wri \<union> wii" and B="wrr" in UnI1)
by (metis r_into_trancl' witness_def)

lemma witnessI4 : "(a, b) \<in> wri \<Longrightarrow> (a, b) \<in> witness"
apply (drule_tac c="(a, b)" and B="wri" and A="wir" in UnI2)
apply (drule_tac c="(a, b)" and A="wir \<union> wri" and B="wii" in UnI1)
apply (drule_tac c="(a, b)" and A="wir \<union> wri \<union> wii" and B="wrr" in UnI1)
by (metis r_into_trancl' witness_def)

lemma lem_witness_irreflexive : "irrefl witness"
apply (unfold irrefl_def)
apply (rule allI)
apply (rule notI)
apply (drule witnessD)
apply (erule disjE)
apply (metis irrefl_def pre_irreflexive)
by (metis def_value lem_pop_then_push_exists lem_push_then_pop_exists lem_wir_then_not_wri lem_wri_then_pop_and_push val_order_def)

lemma lem_witness_antisymmetric : "antisym witness"
by (metis antisymI irrefl_def lem_witness_irreflexive trancl_trans witness_def)

lemma lem_witness_wellformed : "wellformed_history witness val f"
apply (rule wellformed_historyI)
apply (metis trans_trancl witness_def)
apply (rule lem_witness_antisymmetric)
apply (rule lem_witness_irreflexive)
apply (rule operations_countable)
apply (rule operations_finite)
apply (rule def_push_once)
apply (rule def_pop_once)
apply (rule def_value)
apply (metis is_popD is_pushD lem_push_or_pop1 lem_push_or_pop2 witnessD)
by (metis def_val_then_not_pr def_value is_push_def lem_wri val_order_def val_then_not_pr_def val_then_pr witnessD)

 lemma lem_witness_pop_ordered : "is_pop a \<Longrightarrow> is_pop b \<Longrightarrow> a\<noteq>b \<Longrightarrow>(a, b) \<in> witness \<or> (b, a) \<in> witness"
by (metis lem_pr_in_witness pop_ordered subsetD)

lemma lem_witness_push_pop_ordered : "is_push a \<Longrightarrow> is_pop b \<Longrightarrow> (a, b) \<in> witness \<or> (b, a) \<in> witness"
  apply (frule lem_wir_wri_total)
  apply (assumption)
  apply (erule disjE)
  apply (rule disjI1)
  apply (unfold witness_def)
  apply (rule r_into_trancl)
  apply (metis UnCI)
  apply (rule disjI2)
  apply (rule r_into_trancl)
  by (metis UnCI)

lemma lem_pr_wri_ooo: "(ia, rc) \<in> pre \<Longrightarrow> (rc, ib) \<in> wri \<Longrightarrow> (ia, ra)\<in>val \<Longrightarrow> (ib, rb) \<in> val \<Longrightarrow> (ib, ra) \<in> wir \<Longrightarrow> (rb, ra) \<in> pre"
apply (frule_tac a="ia" and b="ra" in lem_val_then_push_and_pop)
apply (frule_tac a="ib" and b="rb" in lem_val_then_push_and_pop)
apply (case_tac "ra=rb")
apply (metis def_push_once lem_wri push_once_def)
apply (erule conjE)+
apply (frule_tac a="ib" and b="ra" in lem_wir_wri_total)
apply (assumption)
apply (erule disjE)
apply (rotate_tac 1)
apply (frule lem_wri_then_pop_and_push)
apply (erule conjE)
apply (erule contrapos_pp)
apply (rule lem_wir_then_not_wri)
apply (assumption)+
apply (metis lem_wri pop_ordered wriI)
by (metis wriD)

lemma lem_wir_wri_ooo: "(ia, rc) \<in> wir \<Longrightarrow> (rc, ib) \<in> wri \<Longrightarrow> (ia, ra)\<in>val \<Longrightarrow> (ib, rb) \<in> val \<Longrightarrow> (ib, ra) \<in> wir \<Longrightarrow> (rb, ra) \<in> pre"
apply (frule_tac a="ia" and b="ra" in lem_val_then_push_and_pop)
apply (frule_tac a="ib" and b="rb" in lem_val_then_push_and_pop)
apply (case_tac "ra=rb")
apply (metis def_push_once push_once_def wriD)
apply (erule conjE)+
by (metis ir_push_and_pop1 lem_wir_then_ir lem_wir_then_not_wri pop_ordered wir_left_right_rec)

lemma lem_witness_ooo : "(ia, ra)\<in>val \<Longrightarrow>(ib, rb)\<in>val \<Longrightarrow> (ia, ib)\<in>witness \<Longrightarrow> (ib, ra)\<in>witness \<Longrightarrow> (rb, ra)\<in>witness"
apply (frule_tac a="ia" and b="ra" in lem_val_then_push_and_pop)
apply (frule_tac a="ib" and b="rb" in lem_val_then_push_and_pop)
apply (erule conjE)+
apply (drule witnessD)+
apply (rule witnessI1)
apply (erule disjE)+
apply (metis ins_def ipr_ooo lem_pr_then_ins lem_push_pr_pop_then_ir)
apply (metis ins_def def_value ipr_ooo is_push_def lem_pr_then_ins lem_wir_then_ir val_order_def)
apply (erule disjE)+

apply (metis lem_wir_wri_ooo wir.wir_intro)

apply (metis lem_irreflexive_then_not_equal lem_witness_irreflexive witnessI3 witnessI4 wriI)
apply (erule disjE)+
apply (metis def_value is_push_def val_order_def)
apply (erule disjE)+
apply (metis lem_wir_wri_ooo)
apply (metis def_value is_push_def val_order_def)
by (metis def_value lem_pop_then_push_exists val_order_def)

theorem insert_pr_relaxed_stack_linearizable : "\<exists> prt .( pre \<subseteq> prt \<and> insert_relaxed_stack prt val f)"
apply (unfold insert_relaxed_stack_def)
apply (unfold insert_relaxed_stack_axioms_def)
apply (rule_tac x="witness" in exI)
apply (rule conjI)
apply (rule lem_pr_in_witness)
apply (rule conjI)
apply (rule lem_witness_wellformed)
apply (rule conjI)
apply (metis lem_pr_in_witness pop_ordered subsetD)
apply (rule conjI)
apply (metis lem_witness_push_pop_ordered)
by (metis lem_witness_ooo)

end
end
