theory insert_relaxed_stack
imports Main concur_history_relations
begin

(* This is the locale for my first linearizability proof. I assume that all
   pops are already ordered and also the order between pushes and pops is known.
   These two preconditions are defined in pop_ordered and push_pop_ordered.
   Some pairs of pushes may be unordered. The property ir_ooo defines what
   such a concurrent history has to provide to allow to show linearizability *)
locale insert_relaxed_stack = wellformed_history +
  assumes pop_ordered: "is_pop a \<Longrightarrow> is_pop b \<Longrightarrow>a\<noteq>b\<Longrightarrow>(a, b) \<in> pre \<or> (b,a)\<in>pre"
  assumes push_pop_ordered: "is_push a \<Longrightarrow>is_pop b \<Longrightarrow> (a, b) : pre \<or> (b,a)\<in>pre"
  assumes ir_ooo : "(ia, ra) \<in> val \<Longrightarrow> (ib, rb) \<in> val \<Longrightarrow> (ia, ib) \<in> pre \<Longrightarrow> (ib, ra) \<in> pre \<Longrightarrow> (rb, ra) \<in> pre"

context insert_relaxed_stack
begin

definition witness :: "'a rel"
where
  "witness = {(a, b). (a, b) \<in> pre \<or> 
           (\<exists>ra rb. (a, ra) \<in> val \<and> (b, rb) \<in> val \<and> ((ra, b)\<in>pre \<or> ((a, rb) \<in> pre \<and> (b, ra) \<in> pre \<and> (rb, ra)\<in>pre)))}"

lemma witnessI1 : "(a, b) \<in> pre \<Longrightarrow> (a, b) \<in> witness"
apply (unfold witness_def)
by (simp add:set_eq_iff)

lemma witnessI2 : "(a, ra) \<in> val \<Longrightarrow> (b, rb) \<in> val \<Longrightarrow> (ra, b) \<in> pre \<Longrightarrow> (a, b) \<in> witness"
apply (unfold witness_def)
apply (simp add:set_eq_iff)
by metis

lemma witnessI3 : "(a, ra) \<in> val \<Longrightarrow> (b, rb) \<in> val \<Longrightarrow> (a, rb) \<in> pre \<Longrightarrow> (b, ra) \<in> pre \<Longrightarrow> (rb, ra) \<in> pre \<Longrightarrow> (a, b) \<in> witness"
apply (unfold witness_def)
apply (simp add:set_eq_iff)
by metis

lemma witnessD : "(a, b) \<in> witness \<Longrightarrow> (a, b) \<in> pre \<or>
 (\<exists>ra rb. (a, ra) \<in> val \<and> (b, rb) \<in> val \<and> (ra, b)\<in>pre) \<or>
 (\<exists>ra rb. (a, ra) \<in> val \<and> (b, rb) \<in> val \<and> (a, rb) \<in> pre \<and> (b, ra) \<in> pre \<and> (rb, ra)\<in>pre)"
apply (unfold witness_def)
apply (simp add:set_eq_iff)
by metis

lemma lem_witness_irreflexive : "irrefl witness"
apply (unfold irrefl_def)
apply (rule allI)
apply (rule notI)
apply (drule witnessD)
apply (erule disjE)
apply (metis lem_irreflexive_then_not_equal pre_irreflexive)

apply (erule disjE)
apply (metis def_val_then_not_pr val_then_not_pr_def)
by (metis def_pop_once lem_irreflexive_then_not_equal pop_once_def pre_irreflexive)

lemma lem_witness_transitive : "trans witness"
apply (insert pre_transitive)
apply (insert pre_irreflexive)
apply (insert pre_antisymmetric)
apply (rule transI)
apply (drule witnessD)
apply (drule witnessD)
apply (erule disjE)+
apply (metis transE witnessI1)
apply (erule disjE)
apply (metis def_val_then_not_pr lem_val_then_push_and_pop push_pop_ordered transE val_then_not_pr_def witnessI1)
apply (erule exE)+
apply (erule conjE)+
apply (frule lem_push_or_pop1)
apply (erule disjE)
apply (smt ir_ooo irrefl_def is_push_def lem_val_then_push_and_pop push_pop_ordered trans_def witnessI2 witnessI3)
apply (smt ir_ooo irrefl_def is_push_def push_pop_ordered trans_def witnessI1)
apply (erule disjE)+
apply (metis def_val_then_not_pr lem_val_then_push_and_pop push_pop_ordered transD val_then_not_pr_def witnessI1)
apply (frule lem_push_or_pop2)
apply (erule disjE)
apply (erule exE)+
apply (erule conjE)+
apply (frule lem_push_then_pop_exists)
apply (erule exE)
apply (erule conjE)
apply (rename_tac rx ry rz)
apply (case_tac "rx=rz")
apply (metis antisym_def def_push_once ir_ooo lem_irreflexive_then_not_equal push_once_def)
apply (frule_tac ia="x" and ra="rx" in is_popI)
apply (frule_tac a="rx" and b="rz" in pop_ordered)
apply (assumption)+
apply (erule disjE)
apply (smt antisymD ir_ooo lem_val_then_push_and_pop push_pop_ordered transD witnessI1)
apply (frule_tac ia="y" and ra="ry" in is_popI)
apply (case_tac "ry=rz")
apply (metis def_push_once lem_irreflexive_then_not_equal push_once_def)
apply (frule_tac a="ry" and b="rz" in pop_ordered)
apply (assumption)+
apply (erule disjE)
apply (smt push_pop_ordered transD witnessI2 witnessI3)
apply (frule_tac ia="x" and ra="rx" in is_pushI)
apply (frule_tac a="x" and b="rz" in push_pop_ordered)
apply (assumption)
apply (erule disjE)
apply (metis push_pop_ordered witnessI2 witnessI3)
apply (smt ir_ooo lem_irreflexive_then_not_equal lem_witness_irreflexive push_pop_ordered trans_def witnessI2)
apply (smt ir_ooo is_pushI lem_irreflexive_then_not_equal push_pop_ordered transD witnessI1)
apply (erule disjE)+
apply (metis def_val_then_not_pr lem_val_then_push_and_pop push_pop_ordered transD val_then_not_pr_def witnessI1)
apply (smt def_val_then_not_pr ir_ooo lem_irreflexive_then_not_equal lem_val_then_push_and_pop push_pop_ordered transD val_then_not_pr_def witnessI1)
apply (erule disjE)
apply (metis def_pop_once pop_once_def transD witnessI1)
apply (erule exE)+
apply (erule conjE)+
apply (rename_tac rx ry ryy rz)
apply (frule_tac ia="x" and ra="rx" in is_pushI)
apply (frule_tac a="x" and b="rz" in push_pop_ordered)
apply (metis is_pop_def)
apply (erule disjE)
apply (smt def_pop_once pop_once_def transE witnessI3)
by (smt ir_ooo lem_irreflexive_then_not_equal transD)

lemma lem_witness_antisymmetric : "antisym witness"
apply (insert pre_transitive)
apply (insert pre_irreflexive)
apply (insert pre_antisymmetric)
apply (rule antisymI)
apply (drule witnessD)+
apply (erule disjE)+
apply (metis antisymD)
apply (erule disjE)
apply (metis def_val_then_not_pr transD val_then_not_pr_def)
apply (metis antisymD ir_ooo lem_irreflexive_then_not_equal)
apply (erule disjE)+
apply (metis def_val_then_not_pr transD val_then_not_pr_def)
apply (metis antisymD ir_ooo lem_irreflexive_then_not_equal)
apply (erule disjE)+
apply (metis def_val_then_not_pr lem_val_then_push_and_pop push_pop_ordered transE val_then_not_pr_def)
apply (metis antisymD def_pop_once def_value pop_once_def val_order_def)
apply (erule disjE)
apply (metis antisym_def def_pop_once def_value pop_once_def val_order_def)
by (metis antisym_def def_pop_once irrefl_def pop_once_def)

lemma lem_pr_in_witness : "pre \<subseteq> witness"
by (metis subrelI witnessI1)

lemma lem_witness_total_on_push : "is_push a \<Longrightarrow> is_push b \<Longrightarrow> a \<noteq> b \<Longrightarrow> (a, b) \<in> witness \<or> (b, a) \<in> witness"
apply (insert pre_transitive)
apply (insert pre_irreflexive)
apply (insert pre_antisymmetric)
apply (frule_tac ia="a" in is_pushD)
apply (frule_tac ia="b" in is_pushD)
apply (erule exE)+
apply (rename_tac rb)
apply (frule_tac a="a" and b="rb" in push_pop_ordered)
apply (metis is_pop_def)
apply (frule_tac a="b" and b="ra" in push_pop_ordered)
apply (metis is_pop_def)
apply (erule disjE)+
apply (frule_tac ia="a" and ra="ra" in is_popI)
apply (frule_tac ia="b" and ra="rb" in is_popI)
apply (case_tac "ra=rb")
apply (metis def_push_once push_once_def)
apply (metis pop_ordered witnessI3)
apply (metis witnessI2)
by (metis witnessI2)

lemma lem_witness_total_on_pop : "is_pop a \<Longrightarrow> is_pop b \<Longrightarrow> a \<noteq> b \<Longrightarrow> (a, b) \<in> witness \<or> (b, a) \<in> witness"
by (metis pop_ordered witnessI1)

lemma lem_witness_total_on_push_pop : "is_push a \<Longrightarrow> is_pop b \<Longrightarrow> (a, b) \<in> witness \<or> (b, a) \<in> witness"
by (metis push_pop_ordered witnessI1)

lemma lem_val_then_pr: "(ia, ra)\<in>val \<Longrightarrow> (ia, ra) \<in> pre"
by (metis def_val_then_not_pr is_pop_def is_push_def push_pop_ordered val_then_not_pr_def)

lemma lem_val_then_not_witness : "(ia, ra) \<in> val \<Longrightarrow> (ra, ia) \<notin> witness"
by (metis def_val_then_not_pr def_value val_order_def val_then_not_pr_def witnessD)

lemma lem_witness_wellformed : "wellformed_history witness val f"
apply (rule wellformed_historyI)
apply (rule lem_witness_transitive)
apply (rule lem_witness_antisymmetric)
apply (rule lem_witness_irreflexive)
apply (rule operations_countable)
apply (rule operations_finite)
apply (rule def_push_once)
apply (rule def_pop_once)
apply (rule def_value)
apply (metis def_push_or_pop witnessD)
by (metis lem_val_then_not_witness val_then_not_prI)

lemma lem_witness_sequential_history : "sequential_set witness val f"
apply (unfold sequential_set_def)
apply (rule conjI)
apply (metis lem_witness_wellformed)
apply (unfold sequential_set_axioms_def)
apply (rule conjI)
apply (metis lem_witness_total_on_push)
apply (rule conjI)
apply (metis lem_witness_total_on_pop)
by (metis lem_witness_total_on_push_pop)

lemma push_witness_pop_then_pr : "(a, b) \<in> witness \<Longrightarrow> is_push a \<Longrightarrow> is_pop b \<Longrightarrow> (a, b) \<in> pre"
by (metis is_pushI lem_irreflexive_then_not_equal lem_witness_irreflexive push_pop_ordered witnessD witnessI1)

lemma lem_witness_ooo : "sequential_lifo witness val"
apply (unfold sequential_lifo_def)
apply (rule allI)+
apply (rule impI)+
apply (drule_tac a="ib" and b="ra" in push_witness_pop_then_pr)
apply (metis is_pushI)
apply (metis is_pop_def)
apply (drule witnessD)
apply (erule disjE)
apply (metis ir_ooo witnessI1)
apply (erule disjE)
apply (metis pre_antisymmetric antisym_def def_pop_once def_value pop_once_def val_order_def)
by (metis def_pop_once pop_once_def witnessI1)

lemma lem_witness_sequential_stack : "sequential_stack witness val f"
apply (unfold sequential_stack_def)
apply (rule conjI)
apply (metis lem_witness_sequential_history)
apply (unfold sequential_stack_axioms_def)
by (metis lem_witness_ooo)

theorem insert_relaxed_stack_linearizable : "\<exists> prt .( pre \<subseteq> prt \<and> sequential_stack prt val f)"
apply (rule_tac x="witness" in exI)
apply (rule conjI)
apply (metis lem_pr_in_witness)
by (metis lem_witness_sequential_stack)

end

end
